/*
 * :set tabstop=4
 * efs.c
 * spd@daphne.cps.unizar.es
 * Tue May  6 00:13:25 DST 1997
 *
 * A program which tries to read EFS by accessing the raw disk device
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#ifdef WIN32
#	include <io.h>
#	include <ctype.h>

#	include <windows.h>

extern int optind;
extern char getopt( int argc, char  *argv[], char *format );

#	include "irix/types.h"
#	include "irix/param.h"
#	define	BITSPERBYTE	8

#	include "irix/dvh.h"			/* volume header information */
#	include "irix/efs_ino.h"
#	include "irix/efs_fs.h"
#	include "irix/efs_sb.h"
#	include "irix/efs_dir.h"

#   define PATH_SEP        '\\'
#else
#	include <sys/time.h>
#	include <sys/param.h>		/* DEV_BSIZE */
#	include <values.h>			/* BITSPERBYTE */

#	include <sys/dvh.h>			/* volume header information */
#	include <sys/fs/efs_ino.h>
#	include <sys/fs/efs_fs.h>
#	include <sys/fs/efs_sb.h>
#	include <sys/fs/efs_dir.h>
#	include <unistd.h>
#endif

#include <errno.h>
#include <time.h>

#include <assert.h>

#if defined(__i386__) || defined (M_I86) || defined (_M_IX86)
#	define	i386
#	define	LITTLEENDIAN
#endif


#define swap_word(a) ( ((a) << 24) | \
	(((a) << 8) & 0x00ff0000) | \
	(((a) >> 8) & 0x0000ff00) | \
	((unsigned long)(a) >>24) )

#define swap_half(a) ( ((a & 0xff) << 8) | ((unsigned short)(a) >> 8) )

#define SWAP_WORD(a) { a = swap_word(a); }
#define SWAP_HALF(a) { a = swap_half(a); }


static char __ident[] = "@(#)(c)SPDsoft, Sun May 25, 1997";
#define VERS_STR        ((char*)&__ident[4])

#ifndef PATH_SEP
#	define PATH_SEP        '/'
#endif
#ifndef SEEK_SET
#	define SEEK_SET 0
#endif

#ifndef MIN
#	define MIN(a,b) ((a)<(b)?(a):(b))
#endif

#ifndef WIN32
#	define	DISK_T	int
#	define	OFF_T	off64_t
#	define	LSEEK(f, o, w)	lseek64(f, o, w)
#	define	READ(file, buff, nbyte) read(file, buff, nbyte)
#	define	CLOSE close
#else
DWORD		_n_read;

#	define	DISK_T	HANDLE
#	define	OFF_T	off_t
#	define	LSEEK(f, o, w) \
	((SetFilePointer(f, o, NULL, FILE_BEGIN) == 0xFFFFFFFF) ? \
	-1 : o )
#	define	READ(file, buff, nbyte) \
	(ReadFile( file, buff, nbyte, (LPDWORD)&_n_read, NULL ) ? \
	_n_read : -1 )
#	define	CLOSE CloseHandle
#endif

char	*app;
int		debug=0;
int		raw=0;
#ifdef WIN32
int		binary=0;
#endif


void usage(void);
int		check_vol_header(struct volume_header *vh);
int     check_sblock(struct efs *efs);
int     check_dinode(struct efs_dinode *di);
int		check_dir( struct efs_dinode *di, size_t sz, char *buf);
int read_dinode(
	DISK_T					disk,
	struct volume_header	*vh,
	int						partition,
	struct efs				*efs,
	efs_ino_t				inum,
	unsigned char			buffer[],
	struct efs_dinode		**di
	);

int		read_file(
	DISK_T					disk,
	struct volume_header	*vh,
	int						partition,
	struct efs				*efs,
	struct efs_dinode		*di,
	int						f
	);

int		read_raw(
	DISK_T					disk,
	char					*buffer,
	OFF_T					start,	
	off_t					n	
	);

void usage(void)
{
	fprintf(stderr,
#ifdef WIN32
	"Use: %s [-hrbd] [-p partition] [-i inode] [-o file] disk\n"
#else
	"Use: %s [-hrd] [-p partition] [-i inode] [-o file] disk\n"
#endif
	, app
	);
	exit(1);
}

int check_dir( struct efs_dinode *di, size_t sz, char *buf)
{
	struct efs_dirblk	*dir;
	int					slot;
	struct efs_dent		*ent;
	char				name[EFS_MAXNAMELEN];
	int					error=0;

	for(
		dir = (struct efs_dirblk*)buf;
		dir < (struct efs_dirblk*) (buf + 512 * sz);
		dir ++ )
	{
#ifdef LITTLEENDIAN
		SWAP_HALF(dir->magic);
#endif /* LITTLEENDIAN */

		if (debug)
		{
		fprintf(stderr, "\t\tmagic:     0x%x (should be 0x%x)\n",
			dir->magic, EFS_DIRBLK_MAGIC);
		fprintf(stderr, "\t\tfirstused: %10d\n", dir->firstused);
		fprintf(stderr, "\t\tslots:     %10d\n", dir->slots);
		}

		if (dir->magic != EFS_DIRBLK_MAGIC)
			error++;

		for( slot = 0; slot < dir->slots; slot ++)
		{
			if( dir->space[slot] != 0 )
			{
				ent = EFS_SLOTOFF_TO_DEP(
					dir,
					EFS_REALOFF(dir->space[slot])
					);
#ifdef LITTLEENDIAN
				SWAP_HALF(ent->ud_inum.s[0]);
				SWAP_HALF(ent->ud_inum.s[1]);
#endif /* LITTLEENDIAN */
				strncpy( name, ent->d_name, ent->d_namelen);
				name[ent->d_namelen] = '\0';
				fprintf(
					stderr, "\t\t\t%3d : %5d \"%s\"\n",
					dir->space[slot] << 1,
					EFS_GET_INUM(ent),
/*					ent->ud_inum.s[1],*/
					name
					);
			}
			else
			{
				fprintf(stderr, "\t\t\t***\n");
			}
		}
	}
	return (error);
}

int     check_dinode(struct efs_dinode *di)
{
	int	i;
	int err=0;

#ifdef LITTLEENDIAN
	SWAP_HALF(di->di_mode);
	SWAP_HALF(di->di_nlink);
	SWAP_HALF(di->di_uid);
	SWAP_HALF(di->di_gid);
	SWAP_WORD(di->di_size);
	SWAP_WORD(di->di_atime);
	SWAP_WORD(di->di_mtime);
	SWAP_WORD(di->di_ctime);
	SWAP_WORD(di->di_gen);
	SWAP_HALF(di->di_numextents);

	for ( i = 0; (i<EFS_DIRECTEXTENTS) && (i<di->di_numextents); i++ )
	{
		unsigned int ui;

		ui=(unsigned int)*(unsigned int *) &(di->di_u.di_extents[i]);
		SWAP_WORD(ui);
		di->di_u.di_extents[i].ex_magic = ui >> 24;
		di->di_u.di_extents[i].ex_bn = ( ui & 0x00FFFFFF );

		ui=(unsigned int)*((unsigned int *) &(di->di_u.di_extents[i]) + 1);
		SWAP_WORD(ui);
		di->di_u.di_extents[i].ex_length = ui >> 24;
		di->di_u.di_extents[i].ex_offset = ( ui & 0x00FFFFFF );
	}
#endif /* LITTLEENDIAN */

	if (debug)
	{
fprintf(stderr,"%x\tmode and type of file\n", di->di_mode);

switch (di->di_mode & IFMT)
	{
		case IFIFO:
			fprintf(stderr,"\t\tnamed pipe (fifo)\n");
			break;
		case IFCHR:
			fprintf(stderr,"\t\tcharacter special\n");
			break;
		case IFDIR:
			fprintf(stderr,"\t\tdirectory\n");
			break;
		case IFBLK:
			fprintf(stderr,"\t\tblock special\n");
			break;
		case IFREG:
			fprintf(stderr,"\t\tregular\n");
			break;
		case IFLNK:
			fprintf(stderr,"\t\tsymbolic link\n");
			break;
		case IFSOCK:
			fprintf(stderr,"\t\tsocket\n");
			break;
		default:
			fprintf(stderr,"\t\tunknown\n");
			break;
	}

if (di->di_mode & ISUID)
	fprintf(stderr,"\t\tset user id on execution\n");
if (di->di_mode & ISGID)
	fprintf(stderr,"\t\tset group id on execution\n");
if (di->di_mode & ISVTX)
	fprintf(stderr,"\t\tsave swapped text even after use\n");
fprintf(stderr,"\t\tPermissions: 0%04o\n", di->di_mode & 0x0fff);

fprintf(stderr,"%d\tnumber of links to file\n", di->di_nlink);
fprintf(stderr,"%d\towner's user id\n", di->di_uid);
fprintf(stderr,"%d\towner's group id\n", di->di_gid);
fprintf(stderr,"%d\tnumber of bytes in file\n", di->di_size);

fprintf(stderr,"\ttime last accessed\t%s", ctime((time_t *)&di->di_atime));
fprintf(stderr,"\ttime last modified\t%s", ctime((time_t *)&di->di_mtime));
fprintf(stderr,"\tlast time inode changed\t%s", ctime((time_t *)&di->di_ctime));

fprintf(stderr,"%d\tgeneration number\n", di->di_gen);
fprintf(stderr,"%10d\tnumber of extents\n", di->di_numextents);
fprintf(stderr,"%x\tversion of inode\n", di->di_version);
fprintf(stderr,"%x\tspare - used by AFS\n", di->di_spare);

	for ( i = 0; (i<EFS_DIRECTEXTENTS) && (i<di->di_numextents); i++ )
	{
		fprintf(stderr,"\textent #\t%d\n",i);
		fprintf(stderr,"\t\t%d\tmagic # (MUST BE ZERO)\n",
			di->di_u.di_extents[i].ex_magic);
		fprintf(stderr,"\t\t%d\tbasic block #\n",
			di->di_u.di_extents[i].ex_bn);
		fprintf(stderr,"\t\t%d\tlength of this extent, in bb's\n",
			di->di_u.di_extents[i].ex_length);
		fprintf(stderr,"\t\t%d\tlogical bb offset into file\n",
			di->di_u.di_extents[i].ex_offset);
	}

	}

	for ( i = 0; (i<EFS_DIRECTEXTENTS) && (i<di->di_numextents); i++ )
	{
		if ( 0 != di->di_u.di_extents[i].ex_magic)
			err++;
	}
	return(err);
}

int     check_sblock(struct efs *efs)
{
/*
 * This is the number of longs contained in the region of the
 * superblock that we need to scan in order to compute the checksum.
 */
#define SBLCKSZ 22

	char	fs_fname[ 7];
	char	fs_fpack[ 7];
	int		i;
	long t6;
	long t7;
	long t8 = 0;
	long v1 = 0;
	long *lptr;
	lptr = (long *)efs;

	/*
	 * Checksum by wpaul@ctr.columbia.edu (Bill Paul)
	 */

    /*
	 * We pass over the superblock data 16 bits at a time.
	 * We XOR each set of 16 bits into an accumulator (v1).
	 * We then shift the accumulator one bit to the left and store
	 * the result into a temporary value (t7). We then store
	 * the high bit of the (unshifted) accumulator in the low
	 * bit of a second temporary value (t8). The accumulator
	 * is then set equal to the logical OR of t7 and t8. Repeat
	 * until you run out of bits. At the end, the accumulator
	 * is the checksum.
	 */

    i = SBLCKSZ;
    while(i)
	{
#ifndef LITTLEENDIAN
		/* Process upper 16 bits */
		t6 = (*lptr & 0xFFFF0000) >> 16;
		v1 = v1 ^ t6;
		t7 = v1 << 1;
		t8 = (v1 & 0x80000000) >> 31;
		v1 = t7 | t8;

		/* Process lower 16 bits */
		t6 = *lptr & 0x0000FFFF;
		v1 = v1 ^ t6;
		t7 = v1 << 1;
		t8 = (v1 & 0x80000000) >> 31;
		v1 = t7 | t8;
#else
		long l;

		l=swap_word(*lptr);
		
		/* Process upper 16 bits */
		t6 = (l & 0xFFFF0000) >> 16;
		v1 = v1 ^ t6;
		t7 = v1 << 1;
		t8 = (v1 & 0x80000000) >> 31;
		v1 = t7 | t8;

		/* Process lower 16 bits */
		t6 = l & 0x0000FFFF;
		v1 = v1 ^ t6;
		t7 = v1 << 1;
		t8 = (v1 & 0x80000000) >> 31;
		v1 = t7 | t8;
#endif
		lptr++;
		i--;
	}

#ifdef LITTLEENDIAN
	SWAP_WORD(efs->fs_size);
	SWAP_WORD(efs->fs_firstcg);
	SWAP_WORD(efs->fs_cgfsize);
	SWAP_HALF(efs->fs_cgisize);
	SWAP_HALF(efs->fs_sectors);
	SWAP_HALF(efs->fs_heads);
	SWAP_HALF(efs->fs_ncg);
	SWAP_HALF(efs->fs_dirty);
	SWAP_WORD(efs->fs_time);
	SWAP_WORD(efs->fs_magic);

	SWAP_WORD(efs->fs_bmsize);
	SWAP_WORD(efs->fs_tfree);
	SWAP_WORD(efs->fs_tinode);
	SWAP_WORD(efs->fs_bmblock);
	SWAP_WORD(efs->fs_replsb);
	SWAP_WORD(efs->fs_lastialloc);

	SWAP_WORD(efs->fs_checksum);
#endif /* LITTLEENDIAN */

	if ( debug )
	{
	fprintf( stderr, "\nefs super block info\n");

fprintf(stderr,"%d\t%s\n",efs->fs_size,"size of filesystem, in sectors");
fprintf(stderr,"%d\t%s\n",efs->fs_firstcg,"bb offset to first cg");
fprintf(stderr,"%d\t%s\n",efs->fs_cgfsize,"size of cylinder group in bb's");
fprintf(stderr,"%d\t%s\n",efs->fs_cgisize,"bb's of inodes per cylinder group");
fprintf(stderr,"%d\t%s\n",efs->fs_sectors,"sectors per track");
fprintf(stderr,"%d\t%s\n",efs->fs_heads,"heads per cylinder");
fprintf(stderr,"%d\t%s\n",efs->fs_ncg,"# of cylinder groups in filesystem");
fprintf(stderr,"%d\t%s\n",efs->fs_dirty,"fs needs to be fsck'd");
fprintf(stderr,"%s\t%s","last super-block update:",
ctime((time_t *)&efs->fs_time));
fprintf(stderr,"%x\t%s (should be %x)\n",efs->fs_magic,"magic number",
EFS_MAGIC);
	strncpy( fs_fname, efs->fs_fname, 6);
	fs_fname[ 6] = '\0';
	fprintf( stderr, "file system name: %6s\n", fs_fname);
	strncpy( fs_fpack, efs->fs_fpack, 6);
	fs_fpack[ 6] = '\0';
	fprintf( stderr, "file system pack name: %6s\n", fs_fpack);
fprintf(stderr,"%d\t%s\n",efs->fs_bmsize,"size of bitmap in bytes");
fprintf(stderr,"%d\t%s\n",efs->fs_tfree,"total free data blocks");
fprintf(stderr,"%d\t%s\n",efs->fs_tinode,"total free inodes");
fprintf(stderr,"%d\t%s\n",efs->fs_bmblock,"bitmap location.");
fprintf(stderr,"%d\t%s\n",efs->fs_replsb,"Location of replicated superblock.");
fprintf(stderr,"%d\t%s\n",efs->fs_lastialloc,"last allocated inode");
/* fs_spare[20] "space for expansion - MUST BE ZERO" */
fprintf(stderr,"%x\t%s (calculated %x)\n",efs->fs_checksum,
"checksum of volume portion of fs", v1);


	}
	return (
		(efs->fs_magic == EFS_MAGIC) &&
		(efs->fs_checksum == v1) &&
		(efs->fs_size > 0) &&
		(efs->fs_ncg > 0) &&
		(efs->fs_firstcg > 0) &&
		(efs->fs_cgfsize > 0) &&
		(efs->fs_cgisize > 0) &&
		(efs->fs_tfree > 0) &&
		(efs->fs_tinode > 0)
	);
}

int		check_vol_header(struct volume_header *vh)
{
	short		i;
	int			sum, *word;


	assert ( sizeof(int) == 4 );

	for (
		word = (int *)vh, sum=0;
		word < ((int *)vh + sizeof(*vh)/sizeof(int) -2);
		word ++
	)
#ifdef LITTLEENDIAN
	sum += swap_word(*word);
#else
	sum += *word;
#endif /* LITTLEENDIAN */


#ifdef LITTLEENDIAN
	SWAP_WORD(vh->vh_magic);
	SWAP_HALF(vh->vh_rootpt);
	SWAP_HALF(vh->vh_swappt);
	SWAP_HALF(vh->vh_dp.dp_cyls);
	SWAP_HALF(vh->vh_dp.dp_shd0);
	SWAP_HALF(vh->vh_dp.dp_trks0);
	SWAP_HALF(vh->vh_dp.dp_unused);
	SWAP_HALF(vh->vh_dp.dp_secs);
	SWAP_HALF(vh->vh_dp.dp_secbytes);
	SWAP_HALF(vh->vh_dp.dp_interleave);
	SWAP_WORD(vh->vh_dp.dp_flags);
	SWAP_WORD(vh->vh_dp.dp_datarate);
	SWAP_WORD(vh->vh_dp.dp_nretries);
	SWAP_WORD(vh->vh_dp.dp_mspw);
	SWAP_HALF(vh->vh_dp.dp_xgap1);
	SWAP_HALF(vh->vh_dp.dp_xsync);
	SWAP_HALF(vh->vh_dp.dp_xrdly);
	SWAP_HALF(vh->vh_dp.dp_xgap2);
	SWAP_HALF(vh->vh_dp.dp_xrgate);
	SWAP_HALF(vh->vh_dp.dp_xwcont);
	for(i=0; i<NVDIR; i++)
	{
		SWAP_WORD(vh->vh_vd[i].vd_lbn);
		SWAP_WORD(vh->vh_vd[i].vd_nbytes);
	}
	for(i=0; i<NPARTAB; i++)
	{
		SWAP_WORD(vh->vh_pt[i].pt_nblks);
		SWAP_WORD(vh->vh_pt[i].pt_firstlbn);
		SWAP_WORD(vh->vh_pt[i].pt_type);
	}
	SWAP_WORD(vh->vh_fill);
	SWAP_WORD(vh->vh_csum);
#endif /* LITTLEENDIAN */


	if (debug)
	{
fprintf(stderr,"%x\t%s (should be %x)\n",vh->vh_magic,
"identifies volume header", VHMAGIC);
fprintf(stderr,"%d\t%s\n",vh->vh_rootpt,"root partition number");
fprintf(stderr,"%d\t%s\n",vh->vh_swappt,"swap partition number");
fprintf(stderr,"%s\t%s\n",vh->vh_bootfile,"name of file to boot");
fprintf(stderr,"%d\t%s\n",vh->vh_dp.dp_skew,"spiral addressing skew");
fprintf(stderr,"%d\t%s\n",vh->vh_dp.dp_gap1,"words of 0 before header");
fprintf(stderr,"%d\t%s\n",vh->vh_dp.dp_gap2,
"words of 0 between hdr and data");
fprintf(stderr,"%d\t%s\n",vh->vh_dp.dp_spares_cyl,"dp_spares_cyl");
fprintf(stderr,"%d\t%s\n",vh->vh_dp.dp_cyls,"number of usable cylinders");
fprintf(stderr,"%d\t%s\n",vh->vh_dp.dp_shd0,"starting head vol 0");
fprintf(stderr,"%d\t%s\n",vh->vh_dp.dp_trks0,
"number of tracks / cylinder vol 0");
fprintf(stderr,"%d\t%s\n",vh->vh_dp.dp_ctq_depth,"Depth of CTQ queue");
fprintf(stderr,"%d\t%s\n",vh->vh_dp.dp_cylshi,
"high byte of 24 bits of cylinder count");
fprintf(stderr,"%d\t%s\n",vh->vh_dp.dp_unused,"not used");
fprintf(stderr,"%d\t%s\n",vh->vh_dp.dp_secs,"number of sectors/track");
fprintf(stderr,"%d\t%s (this makes a %ld Mb disk)\n",vh->vh_dp.dp_secbytes,
"length of sector in bytes",
vh->vh_dp.dp_cyls *       /* usable cylinders */
vh->vh_dp.dp_trks0 *      /* tracks/cylinder */
vh->vh_dp.dp_secs         /* sectors/track */
/ 1024 *
vh->vh_dp.dp_secbytes     /* bytes/sector */
/ 1024
);
fprintf(stderr,"%d\t%s\n",vh->vh_dp.dp_interleave,"sector interleave");
fprintf(stderr,"%X\t%s\n",vh->vh_dp.dp_flags,"controller characteristics");

fprintf(stderr,"\t%c\t%s\n",vh->vh_dp.dp_flags & DP_SECTSLIP ? 'y' : 'n' ,
"sector slip to spare sector");
fprintf(stderr,"\t%c\t%s\n",vh->vh_dp.dp_flags & DP_SECTFWD ? 'y' : 'n' ,
"forward to replacement sector"); 
fprintf(stderr,"\t%c\t%s\n",vh->vh_dp.dp_flags & DP_TRKFWD ? 'y' : 'n' ,
"forward to replacement track");
fprintf(stderr,"\t%c\t%s\n",vh->vh_dp.dp_flags & DP_MULTIVOL ? 'y' : 'n' ,
"multiple volumes per spindle");
fprintf(stderr,"\t%c\t%s\n",vh->vh_dp.dp_flags & DP_IGNOREERRORS ? 'y' : 'n' ,
"transfer data regardless of errors");
fprintf(stderr,"\t%c\t%s\n",vh->vh_dp.dp_flags & DP_RESEEK ? 'y' : 'n' ,
"recalibrate as last resort");
fprintf(stderr,"\t%c\t%s\n",vh->vh_dp.dp_flags & DP_CTQ_EN ? 'y' : 'n' ,
"enable command tag queueing");

fprintf(stderr,"%d\t%s\n",vh->vh_dp.dp_datarate,
"bytes/sec for kernel stats");
fprintf(stderr,"%d\t%s\n",vh->vh_dp.dp_nretries,
"max num retries on data error");
fprintf(stderr,"%d\t%s\n",vh->vh_dp.dp_mspw,
"ms per word to xfer, for iostat");
fprintf(stderr,"%d\t%s\n",vh->vh_dp.dp_xgap1,
"Gap 1 for xylogics controllers");
fprintf(stderr,"%d\t%s\n",vh->vh_dp.dp_xsync,
"sync delay for xylogics controllers");
fprintf(stderr,"%d\t%s\n",vh->vh_dp.dp_xrdly,
"read delay for xylogics controllers");
fprintf(stderr,"%d\t%s\n",vh->vh_dp.dp_xgap2,
"gap 2 for xylogics controllers");
fprintf(stderr,"%d\t%s\n",vh->vh_dp.dp_xrgate,
"read gate for xylogics controllers");
fprintf(stderr,"%d\t%s\n",vh->vh_dp.dp_xwcont,
"write continuation for xylogics");

fprintf(stderr,"\t%s\n", "volume_directory");
for(i=0; i<NVDIR; i++)
{
	if (( vh->vh_vd[i].vd_lbn == 0 ) && ( vh->vh_vd[i].vd_nbytes == 0 ))
		break;
	fprintf(stderr,"\t%d\t%s\t%s\n",i,vh->vh_vd[i].vd_name,"name");
	fprintf(stderr,"\t%d\t%d\t%s\n",i,vh->vh_vd[i].vd_lbn,
		"logical block number");
	fprintf(stderr,"\t%d\t%d\t%s\n",i,vh->vh_vd[i].vd_nbytes,
		"file length in bytes");
}

fprintf(stderr,"\t%s\n", "partition_table");
for(i=0; i<NPARTAB; i++)
{
	if ( vh->vh_pt[i].pt_nblks != 0 )
	{
		char *use[]=
		{
			"partition is volume header",
			"partition is used for repl trks",
			"partition is used for repl secs",
			"partition is used for data",
			"partition is 4.2BSD file system",
			"partition is SysV file system",
			"partition is entire volume",
			"partition is sgi EFS",
			"partition is part of a logical vol",
			"part of a \"raw\" logical vol",
			"partition is sgi XFS",
			"partition is sgi XFS log",
			"partition is part of an XLV vol"
		};

		fprintf(stderr,"\tpartition %d\n",i);
		fprintf(stderr,"%d\t%s\n",vh->vh_pt[i].pt_nblks,
		"# of logical blks in partition");
		fprintf(stderr,"%d\t%s\n",vh->vh_pt[i].pt_firstlbn,
		"first lbn of partition");
		fprintf(stderr,"%d\t%s (%s)\n",vh->vh_pt[i].pt_type,
		"use of partition",
		use[vh->vh_pt[i].pt_type]
		);
	}
}

fprintf(stderr,"\n%d\t%s\n",vh->vh_fill,"fill out to 512 bytes");


fprintf(stderr,"0x%X\txchecksum of sector (calculated 0x%X)\n",
vh->vh_csum, -sum);
	}
	return (
		( vh->vh_csum == -sum ) &&
		( vh->vh_magic == VHMAGIC)
	);

}

void main(int argc, char ** argv)
{
	extern char			*optarg;
	extern int 			optind,opterr;
	int 				opt;

	int					err=0;

	DISK_T				disk;
	int					partition=0;
	char				*name=NULL;
	int					f=0;
	OFF_T				start;

	unsigned char			bblock[DEV_BSIZE];
	unsigned char			sblock[DEV_BSIZE]; /* superblock */
	unsigned char			buffer[DEV_BSIZE]; /* inodes */

	struct efs				*efs;
	struct volume_header	*vh;

	struct efs_dinode	*di;
	efs_ino_t			inum = EFS_ROOTINO;

	if ((app = strrchr(argv[0], PATH_SEP))  != NULL)
		app = app+1;
	else
		app = argv[0];

	while ( (opt=getopt(argc,argv,
#ifdef WIN32
		"hrdbp:i:Vo:"
#else
		"hrdp:i:Vo:"
#endif
									)) != EOF )
	{
		switch(opt)
		{
#ifdef WIN32
			case 'b':
				binary=1;
				break;
#endif
			case 'r':
				raw=1;
				break;

			case 'd':
				debug=1;
				break;

			case 'o':
				name=optarg;
				break;

			case 'V':
				fprintf(stderr, "%s: %s\n", app, VERS_STR);
				exit(0);
				break;
			case 'i':
				inum=(efs_ino_t) atol(optarg);
				if (inum<0)
				{
					fprintf(stderr, "%s: invalid inode (%d)\n",
					app, inum);
					exit(1);
				}
				break;
			case 'p':

				partition=atoi(optarg);

				if ((partition<0) || (partition>=NPARTAB))
				{
					fprintf(stderr, "%s: invalid parition (%d)\n",
					app, partition);
					exit(1);
				}
			break;
			case 'h':
			default:
				usage();
				break;
		}
	}

	if ( argv[optind] == NULL )
		usage();


#ifdef WIN32
	if ( !raw && ( 0==strncmp ( "\\\\.\\PHYSICALDRIVE", argv[optind], 17 )))
#else
	if ( !raw && ( 0==strncmp ( "/dev/rdsk/", argv[optind], 10 )))
#endif
	{
		fprintf(stderr,"Warning: assuming raw I/O (%s)\n", argv[optind]);
		raw = 1;
	}	

#ifdef WIN32
	if (INVALID_HANDLE_VALUE == ( disk = CreateFile(
	(LPCTSTR) argv[optind],
	GENERIC_READ,
	0, /* Not shared */
	NULL, /* Not inherited by child processes */
	OPEN_EXISTING,
	FILE_ATTRIBUTE_NORMAL | FILE_FLAG_RANDOM_ACCESS,
	NULL
	)))
#else
	if ( -1 == ( disk = open( argv[optind], O_RDONLY )))
#endif
	{
		perror(argv[optind]);
		err = 1;
	}
	else
	{
		if ( DEV_BSIZE != READ( disk, bblock, DEV_BSIZE ))
		{
			perror("read volume header");
			err = 2;
		}
		else
		{
			vh = (struct volume_header *)bblock;

			if ( !check_vol_header( vh ) )
			{
				fprintf(stderr,"%s: corrupted boot block\n", app);
				err = 3;
			}
			else
			{
				/*
				 * Translate partition starting cylinder to offset from 0
				 */

				start = ( 1 + vh->vh_pt[partition].pt_firstlbn ) * 512 ;

				if ( start != LSEEK(disk, start, SEEK_SET))
				{
					perror("lseek");
					err = 4;
				}
				else
				{
					efs = (struct efs *)sblock;
					if ( sizeof(sblock) != READ( disk, efs, sizeof(sblock)))
					{
						perror("read sblock");
						err = 5;
					}
					else
					{
						if ( !check_sblock(efs) )
						{
							fprintf(stderr,"%s: corrupted sblock\n", app);
							err = 6;
						}
						else
						{
							di = NULL;
							if ( 0 != (err = read_dinode(
								disk,
								vh,
								partition,
								efs,
								inum,
								buffer,
								&di
								)))
							{
								fprintf(stderr,"%s: bad inode\n", app);
								err = 7;
							}
							else
							{
								if ( name != NULL)
								{
									if (
										( name[0]=='-' ) &&
										( name[1]==0x00 )
										)
									{
										f = fileno(stdout);
#ifdef WIN32
										if( -1 == _setmode(
											_fileno( stdout ),
											_O_BINARY ))
										{
						perror("Warning! Can't set stdout mode to binary");
										}
#endif
									}
									else if ( -1 == (f = open(
										name,
										O_TRUNC|O_WRONLY|O_CREAT
#ifdef WIN32
										|O_BINARY
#endif
										,0666)))
									{
										perror("open");
										f=0;
									}
								}
								if (debug)
									fprintf(stderr,"####\tinode data:\n");

								read_file(
									disk,
									vh,
									partition,
									efs,
									di,
									f
									);

								if (f!=0)
								{
#ifndef WIN32
									ftruncate(f, di->di_size);
#else
									_chsize(f, di->di_size);
#endif
									close(f);
								}
								
							}
						}
					}
				}
			}
		}
		CLOSE(disk);
	}
	exit ( err );
}


int read_dinode(
	DISK_T					disk,
	struct volume_header	*vh,
	int						partition,
	struct efs				*efs,
	efs_ino_t				inum,
	unsigned char			buffer[],
	struct efs_dinode		**di
	)
{
	int		cg;
	int		in;
	OFF_T	start;
	OFF_T	remain;

	cg = inum / (efs->fs_cgisize * (512 / sizeof( struct efs_dinode)));
	in = inum % (efs->fs_cgisize * (512 / sizeof( struct efs_dinode)));

	if (debug)
		fprintf(stderr, "inode %2d [%d:%d]\n", inum, cg, in);

	start = 512 * (efs->fs_firstcg + efs->fs_cgfsize * cg)
			+ sizeof( struct efs_dinode) * in +
			vh->vh_pt[partition].pt_firstlbn * 512;


	if ( raw && ( start % DEV_BSIZE != 0 ))
	{
		remain = start % DEV_BSIZE ;
		start -= remain;
	}
	else
	{
		remain = 0L;
	}

	if ( start != LSEEK(disk, start, SEEK_SET))
	{
		perror("lseek inum");
		return(1);
	}

	if ( DEV_BSIZE != READ( disk, buffer, DEV_BSIZE ))
	{
		perror("read inode");
		return(2);
	}

	if ( *di == NULL )
		*di = (struct efs_dinode *)(buffer + remain);
	else
		memcpy( (char *)*di, buffer + remain, sizeof(struct efs_dinode));

	return (check_dinode(*di));
}

int		read_file(
	DISK_T					disk,
	struct volume_header	*vh,
	int						partition,
	struct efs				*efs,
	struct efs_dinode		*di,
	int						f
)
{
	int		i;
	char	buf[512 * 256];

	static unsigned long	written=0;


	if( di->di_numextents <= EFS_DIRECTEXTENTS )
	{

		if (debug)
			fprintf(stderr, "\t%d extents (short)\n", di->di_numextents);

		for( i = 0; i < di->di_numextents; i++ )
		{
			if ( -1 == read_raw( disk, buf,
				512 * ( vh->vh_pt[partition].pt_firstlbn +
				di->di_u.di_extents[i].ex_bn ),
				(512 * di->di_u.di_extents[i].ex_length)))
				return (-1);

if (debug) fprintf(stderr, "Writting %d bytes to %d\n",
512 * di->di_u.di_extents[i].ex_length,
512 * di->di_u.di_extents[i].ex_offset);

printf("%d %d %d\n",
written,
512 * di->di_u.di_extents[i].ex_offset,
512 * di->di_u.di_extents[i].ex_length);

/* not really an assertion, just wanna know if this is true... */
assert ( written == (512 * di->di_u.di_extents[i].ex_offset ));
written += (512 * di->di_u.di_extents[i].ex_length );


			if ( f != 0 )
			{
				if ( f != fileno(stdout))
#ifdef WIN32
					lseek(
#else
					lseek64(
#endif
						f,
						512 * di->di_u.di_extents[i].ex_offset,
						SEEK_SET
						);
					
					if ( -1 == write(
						f,
						buf,
						512 * di->di_u.di_extents[i].ex_length
						))
						perror("write");

			}

			if( di->di_mode & S_IFDIR)
			{
				check_dir( di, di->di_u.di_extents[i].ex_length, buf);
			}
		}
	}
	else /* indirect extents */
	{
		int		num_extents_done = 0;

		if (debug)
			fprintf(stderr, "\t%d extents (long)\n", di->di_numextents);

		for(i = 0; i < di->di_u.di_extents[0].ex_offset; i++)
		{
			int		block;

			if (debug)
				printf( "####\tindirect extent %d\n", i);

			for( block=0; block<di->di_u.di_extents[i].ex_length; block ++)
			{
				extent	extents[512 / sizeof(extent)];
				int		indirect_extent_num;

				if (debug) printf( "########\tblock %d\n", block);

				if ( -1 == read_raw( disk, (char *)extents,
					(512 * (
					vh->vh_pt[partition].pt_firstlbn +
					di->di_u.di_extents[i].ex_bn +
					block
					)), 512)
					)
					return (-1);
#ifdef LITTLEENDIAN
{
				extent *foo = (extent *)extents;
				unsigned int ui;

				while( foo < ((extent *)extents)+ 512 / sizeof(extent) )
				{
					ui=(unsigned int)*((unsigned int *) foo);
					SWAP_WORD(ui);
					foo->ex_magic = ui >> 24;
					foo->ex_bn = ( ui & 0x00FFFFFF );

					ui=(unsigned int)*((unsigned int *) foo + 1);
					SWAP_WORD(ui);
					foo->ex_length = ui >> 24;
					foo->ex_offset = ( ui & 0x00FFFFFF );
					foo++;
				}

}
#endif /* LITTLEENDIAN */

				for(
					indirect_extent_num = 0;
					(indirect_extent_num < (512 / sizeof( extent))) &&
					(num_extents_done < di->di_numextents);
					indirect_extent_num ++, num_extents_done ++
					)
				{
					if (debug) printf( "############\textent %d\n", 
						num_extents_done );

					if ( -1 == read_raw( disk, buf,
						512 * ( vh->vh_pt[partition].pt_firstlbn +
						extents[indirect_extent_num].ex_bn ),
						(512 * extents[indirect_extent_num].ex_length)))
						return (-1);

if (debug) fprintf(stderr, "Writting %d bytes to %d\n",
512 * extents[indirect_extent_num].ex_length,
512 * extents[indirect_extent_num].ex_offset);
assert ( written == (512 * extents[indirect_extent_num].ex_offset));
written += (512 * extents[indirect_extent_num].ex_length );

					if ( f != 0 )
					{
						if ( f != fileno(stdout))
#ifdef WIN32
					lseek(
#else
					lseek64(
#endif
								f,
								512 * extents[indirect_extent_num].ex_offset,
								SEEK_SET
								);
							
							if ( -1 == write(
								f,
								buf,
								512 * extents[indirect_extent_num].ex_length
								))
								perror("write");

					}
					if( di->di_mode & S_IFDIR)
					{
						check_dir( di,
							extents[indirect_extent_num].ex_length, buf);
					}
				}
			}
		}
	}

	return(0);

}

int		read_raw(
	DISK_T					disk,
	char					*buffer,
	OFF_T					start,	
	off_t					len
)
{
	off_t	remain;
	off_t	rlen;

#ifndef WIN32
	if (debug) fprintf(stderr, "### Reading %d bytes from %llu\n", len, start);
#else
	if (debug) fprintf(stderr, "### Reading %d bytes from %d\n", len, start);
#endif



	if ( raw && ( start % DEV_BSIZE != 0 ))
	{
		rlen = len + 512 &~ 511;
		remain = start % DEV_BSIZE ;
		start -= remain;
	}
	else
	{
		rlen = len;
		remain = 0L;
	}

	if ( start != LSEEK(disk, start, SEEK_SET))
	{
		perror("lseek inum");
		return(-1);
	}

	if ( rlen != (len=READ( disk, buffer, rlen )))
	{
		perror("read raw");
		return(-1);
	}

	if ( remain != 0 )
		memmove(buffer, buffer + remain, len);

	return ( len );
}

