/*
 * :set tabstop=4
 * efs_db.c
 * spd@daphne.cps.unizar.es
 * Sun Jun  8 22:24:57 DST 1997
 *
 * EFS debug
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <time.h>

#ifdef WIN32
#	include <io.h>
#	include <ctype.h>

#	include <windows.h>

#	include "irix/types.h"
#	include "irix/param.h"
#	define  BITSPERBYTE 8

#	include "irix/dvh.h"			/* volume header information */
#	include "irix/efs_ino.h"
#	include "irix/efs_fs.h"
#	include "irix/efs_sb.h"
#	include "irix/efs_dir.h"

#	define S_IFIFO     0x1000  /* fifo */
#	define S_IFBLK     0x6000  /* block special */
#	define S_IFSOCK    0xC000  /* socket */
#	define S_IFLNK     0xA000  /* symbolic link */
#	ifndef S_ISUID
#		define S_ISUID     0x800   /* set user id on execution */
#		define S_ISGID     0x400   /* set group id on execution */
#	endif  /* S_ISUID */
#	define S_ISVTX     0x200   /* save swapped text even after use */

#else /* IRIX */

#	include <sys/time.h>
#	include <sys/param.h>			/* DEV_BSIZE */
#	include <values.h>				/* BITSPERBYTE */

#	include <sys/dvh.h>				/* volume header information */
#	include <sys/fs/efs_ino.h>
#	include <sys/fs/efs_fs.h>
#	include <sys/fs/efs_sb.h>
#	include <sys/fs/efs_dir.h>
#	include <unistd.h>

#endif /* WIN32 */

#include "efs_db.h"


void efs_pstat( FILE *f, struct stat *sbp )
{
	int			isdevice=0;

	fprintf(f,"\ttype:");
	switch (sbp->st_mode & S_IFMT)
	{
		case S_IFIFO:
			fprintf(f,"\tnamed pipe (fifo)\n");
			break;
		case S_IFCHR:
			fprintf(f,"\tcharacter special\n");
			isdevice=1;
			break;
		case S_IFDIR:
			fprintf(f,"\tdirectory\n");
			break;
		case S_IFBLK:
			fprintf(f,"\tblock special\n");
			isdevice=1;
			break;
		case S_IFREG:
			fprintf(f,"\tregular\n");
			break;
		case S_IFLNK:
			fprintf(f,"\tsymbolic link\n");
			break;
		case S_IFSOCK:
			fprintf(f,"\tsocket\n");
			break;
		default:
			fprintf(f,"\tunknown\n");
			break;
	}

	if (isdevice)
		fprintf(f, "Device number:%d,%d\n",
				(sbp->st_rdev>>8) & 0xff,
				sbp->st_rdev & 0xff );

	fprintf(f, "Resides on device:%d,%d\n",
			(sbp->st_dev>>8) & 0xff,
			sbp->st_dev & 0xff);

	fprintf(f, "I-node: %d; Links: %d; Size: %ld\n",
			sbp->st_ino, sbp->st_nlink, sbp->st_size);


	fprintf(f, "Owner ID: %d\n", sbp->st_uid );
	fprintf(f, "Group ID: %d\n", sbp->st_gid );

	if ((sbp->st_mode & S_ISUID )== S_ISUID )
		fprintf(f, "Set-user-ID\n");

	if ((sbp->st_mode & S_ISGID )== S_ISGID )
		fprintf(f, "Set-group-ID\n");

	if ((sbp->st_mode & S_ISVTX )== S_ISVTX )
		fprintf(f, "Save swapped text after use\n");

	fprintf(f, "Permissions: %o\n",sbp->st_mode & 0xfff);
	fprintf(f, "Last access:         %s",
		asctime(localtime(&sbp->st_atime )));
	fprintf(f, "Last modification:   %s",
		asctime(localtime(&sbp->st_mtime )));
	fprintf(f, "Last status change:  %s",
		asctime(localtime(&sbp->st_ctime )));

}

void efs_pdinode(FILE *f, struct efs_dinode *di)
{
	int	i;

	fprintf(stderr,"%x\tmode and type of file\n", di->di_mode);

	switch (di->di_mode & IFMT)
	{
		case IFIFO:
			fprintf(stderr,"\t\tnamed pipe (fifo)\n");
			break;
		case IFCHR:
			fprintf(stderr,"\t\tcharacter special\n");
			break;
		case IFDIR:
			fprintf(stderr,"\t\tdirectory\n");
			break;
		case IFBLK:
			fprintf(stderr,"\t\tblock special\n");
			break;
		case IFREG:
			fprintf(stderr,"\t\tregular\n");
			break;
		case IFLNK:
			fprintf(stderr,"\t\tsymbolic link\n");
			break;
		case IFSOCK:
			fprintf(stderr,"\t\tsocket\n");
			break;
		default:
			fprintf(stderr,"\t\tunknown\n");
			break;
	}

if (di->di_mode & ISUID)
	fprintf(stderr,"\t\tset user id on execution\n");
if (di->di_mode & ISGID)
	fprintf(stderr,"\t\tset group id on execution\n");
if (di->di_mode & ISVTX)
	fprintf(stderr,"\t\tsave swapped text even after use\n");
fprintf(stderr,"\t\tPermissions: 0%04o\n", di->di_mode & 0x0fff);

fprintf(stderr,"%d\tnumber of links to file\n", di->di_nlink);
fprintf(stderr,"%d\towner's user id\n", di->di_uid);
fprintf(stderr,"%d\towner's group id\n", di->di_gid);
fprintf(stderr,"%d\tnumber of bytes in file\n", di->di_size);

fprintf(stderr,"\ttime last accessed\t%s", ctime((time_t *)&di->di_atime));
fprintf(stderr,"\ttime last modified\t%s", ctime((time_t *)&di->di_mtime));
fprintf(stderr,"\tlast time inode changed\t%s", ctime((time_t *)&di->di_ctime));

fprintf(stderr,"%d\tgeneration number\n", di->di_gen);
fprintf(stderr,"%10d\tnumber of extents\n", di->di_numextents);
fprintf(stderr,"%x\tversion of inode\n", di->di_version);
fprintf(stderr,"%x\tspare - used by AFS\n", di->di_spare);

	for ( i = 0; (i<EFS_DIRECTEXTENTS) && (i<di->di_numextents); i++ )
	{
		fprintf(stderr,"\textent #\t%d\n",i);
		fprintf(stderr,"\t\t%d\tmagic # (MUST BE ZERO)\n",
			di->di_u.di_extents[i].ex_magic);
		fprintf(stderr,"\t\t%d\tbasic block #\n",
			di->di_u.di_extents[i].ex_bn);
		fprintf(stderr,"\t\t%d\tlength of this extent, in bb's\n",
			di->di_u.di_extents[i].ex_length);
		fprintf(stderr,"\t\t%d\tlogical bb offset into file\n",
			di->di_u.di_extents[i].ex_offset);
	}
}

void efs_pvh(FILE *f, struct volume_header *vh)
{
	short		i;

fprintf(stderr,"%x\t%s (should be %x)\n",vh->vh_magic,
"identifies volume header", VHMAGIC);
fprintf(stderr,"%d\t%s\n",vh->vh_rootpt,"root partition number");
fprintf(stderr,"%d\t%s\n",vh->vh_swappt,"swap partition number");
fprintf(stderr,"%s\t%s\n",vh->vh_bootfile,"name of file to boot");
fprintf(stderr,"%d\t%s\n",vh->vh_dp.dp_skew,"spiral addressing skew");
fprintf(stderr,"%d\t%s\n",vh->vh_dp.dp_gap1,"words of 0 before header");
fprintf(stderr,"%d\t%s\n",vh->vh_dp.dp_gap2,
"words of 0 between hdr and data");
fprintf(stderr,"%d\t%s\n",vh->vh_dp.dp_spares_cyl,"dp_spares_cyl");
fprintf(stderr,"%d\t%s\n",vh->vh_dp.dp_cyls,"number of usable cylinders");
fprintf(stderr,"%d\t%s\n",vh->vh_dp.dp_shd0,"starting head vol 0");
fprintf(stderr,"%d\t%s\n",vh->vh_dp.dp_trks0,
"number of tracks / cylinder vol 0");
fprintf(stderr,"%d\t%s\n",vh->vh_dp.dp_ctq_depth,"Depth of CTQ queue");
#ifndef IRIX5
fprintf(stderr,"%d\t%s\n",vh->vh_dp.dp_cylshi,
"high byte of 24 bits of cylinder count");
fprintf(stderr,"%d\t%s\n",vh->vh_dp.dp_unused,"not used");
#endif
fprintf(stderr,"%d\t%s\n",vh->vh_dp.dp_secs,"number of sectors/track");
fprintf(stderr,"%d\t%s (this makes a %ld Mb disk)\n",vh->vh_dp.dp_secbytes,
"length of sector in bytes",
vh->vh_dp.dp_cyls *       /* usable cylinders */
vh->vh_dp.dp_trks0 *      /* tracks/cylinder */
vh->vh_dp.dp_secs         /* sectors/track */
/ 1024 *
vh->vh_dp.dp_secbytes     /* bytes/sector */
/ 1024
);
fprintf(stderr,"%d\t%s\n",vh->vh_dp.dp_interleave,"sector interleave");
fprintf(stderr,"%X\t%s\n",vh->vh_dp.dp_flags,"controller characteristics");

fprintf(stderr,"\t%c\t%s\n",vh->vh_dp.dp_flags & DP_SECTSLIP ? 'y' : 'n' ,
"sector slip to spare sector");
fprintf(stderr,"\t%c\t%s\n",vh->vh_dp.dp_flags & DP_SECTFWD ? 'y' : 'n' ,
"forward to replacement sector"); 
fprintf(stderr,"\t%c\t%s\n",vh->vh_dp.dp_flags & DP_TRKFWD ? 'y' : 'n' ,
"forward to replacement track");
fprintf(stderr,"\t%c\t%s\n",vh->vh_dp.dp_flags & DP_MULTIVOL ? 'y' : 'n' ,
"multiple volumes per spindle");
fprintf(stderr,"\t%c\t%s\n",vh->vh_dp.dp_flags & DP_IGNOREERRORS ? 'y' : 'n' ,
"transfer data regardless of errors");
fprintf(stderr,"\t%c\t%s\n",vh->vh_dp.dp_flags & DP_RESEEK ? 'y' : 'n' ,
"recalibrate as last resort");
fprintf(stderr,"\t%c\t%s\n",vh->vh_dp.dp_flags & DP_CTQ_EN ? 'y' : 'n' ,
"enable command tag queueing");

fprintf(stderr,"%d\t%s\n",vh->vh_dp.dp_datarate,
"bytes/sec for kernel stats");
fprintf(stderr,"%d\t%s\n",vh->vh_dp.dp_nretries,
"max num retries on data error");
fprintf(stderr,"%d\t%s\n",vh->vh_dp.dp_mspw,
"ms per word to xfer, for iostat");
fprintf(stderr,"%d\t%s\n",vh->vh_dp.dp_xgap1,
"Gap 1 for xylogics controllers");
fprintf(stderr,"%d\t%s\n",vh->vh_dp.dp_xsync,
"sync delay for xylogics controllers");
fprintf(stderr,"%d\t%s\n",vh->vh_dp.dp_xrdly,
"read delay for xylogics controllers");
fprintf(stderr,"%d\t%s\n",vh->vh_dp.dp_xgap2,
"gap 2 for xylogics controllers");
fprintf(stderr,"%d\t%s\n",vh->vh_dp.dp_xrgate,
"read gate for xylogics controllers");
fprintf(stderr,"%d\t%s\n",vh->vh_dp.dp_xwcont,
"write continuation for xylogics");

fprintf(stderr,"\t%s\n", "volume_directory");
for(i=0; i<NVDIR; i++)
{
	if (( vh->vh_vd[i].vd_lbn == 0 ) && ( vh->vh_vd[i].vd_nbytes == 0 ))
		break;
	fprintf(stderr,"\t%d\t%s\t%s\n",i,vh->vh_vd[i].vd_name,"name");
	fprintf(stderr,"\t%d\t%d\t%s\n",i,vh->vh_vd[i].vd_lbn,
		"logical block number");
	fprintf(stderr,"\t%d\t%d\t%s\n",i,vh->vh_vd[i].vd_nbytes,
		"file length in bytes");
}

fprintf(stderr,"\t%s\n", "partition_table");
for(i=0; i<NPARTAB; i++)
{
	if ( vh->vh_pt[i].pt_nblks != 0 )
	{
		char *use[]=
		{
			"partition is volume header",
			"partition is used for repl trks",
			"partition is used for repl secs",
			"partition is used for data",
			"partition is 4.2BSD file system",
			"partition is SysV file system",
			"partition is entire volume",
			"partition is sgi EFS",
			"partition is part of a logical vol",
			"part of a \"raw\" logical vol",
			"partition is sgi XFS",
			"partition is sgi XFS log",
			"partition is part of an XLV vol"
		};

		fprintf(stderr,"\tpartition %d\n",i);
		fprintf(stderr,"%d\t%s\n",vh->vh_pt[i].pt_nblks,
		"# of logical blks in partition");
		fprintf(stderr,"%d\t%s\n",vh->vh_pt[i].pt_firstlbn,
		"first lbn of partition");
		fprintf(stderr,"%d\t%s (%s)\n",vh->vh_pt[i].pt_type,
		"use of partition",
		use[vh->vh_pt[i].pt_type]
		);
	}
}

fprintf(stderr,"\n%d\t%s\n",vh->vh_fill,"fill out to 512 bytes");


fprintf(stderr,"0x%X\txchecksum of sector\n",
vh->vh_csum);
}

void efs_psblock(FILE *f, struct efs *efs)
{
	char	fs_fname[ 7];
	char	fs_fpack[ 7];
	
fprintf(stderr,"%d\t%s\n",efs->fs_size,"size of filesystem, in sectors");
fprintf(stderr,"%d\t%s\n",efs->fs_firstcg,"bb offset to first cg");
fprintf(stderr,"%d\t%s\n",efs->fs_cgfsize,"size of cylinder group in bb's");
fprintf(stderr,"%d\t%s\n",efs->fs_cgisize,"bb's of inodes per cylinder group");
fprintf(stderr,"%d\t%s\n",efs->fs_sectors,"sectors per track");
fprintf(stderr,"%d\t%s\n",efs->fs_heads,"heads per cylinder");
fprintf(stderr,"%d\t%s\n",efs->fs_ncg,"# of cylinder groups in filesystem");
fprintf(stderr,"%d\t%s\n",efs->fs_dirty,"fs needs to be fsck'd");
fprintf(stderr,"%s\t%s","last super-block update:",
ctime((time_t *)&efs->fs_time));
fprintf(stderr,"%x\t%s (should be %x)\n",efs->fs_magic,"magic number",
EFS_MAGIC);
	strncpy( fs_fname, efs->fs_fname, 6);
	fs_fname[ 6] = '\0';
	fprintf( stderr, "file system name: %6s\n", fs_fname);
	strncpy( fs_fpack, efs->fs_fpack, 6);
	fs_fpack[ 6] = '\0';
	fprintf( stderr, "file system pack name: %6s\n", fs_fpack);
fprintf(stderr,"%d\t%s\n",efs->fs_bmsize,"size of bitmap in bytes");
fprintf(stderr,"%d\t%s\n",efs->fs_tfree,"total free data blocks");
fprintf(stderr,"%d\t%s\n",efs->fs_tinode,"total free inodes");
fprintf(stderr,"%d\t%s\n",efs->fs_bmblock,"bitmap location.");
fprintf(stderr,"%d\t%s\n",efs->fs_replsb,"Location of replicated superblock.");
fprintf(stderr,"%d\t%s\n",efs->fs_lastialloc,"last allocated inode");
/* fs_spare[20] "space for expansion - MUST BE ZERO" */
fprintf(stderr,"%x\t%s\n",efs->fs_checksum,
"checksum of volume portion of fs");
}

