/*
 * :set tabstop=4
 * efs_io.c
 * spd@daphne.cps.unizar.es
 * Sun Jun  8 01:17:16 DST 1997
 *
 * IRIX 6.x and WinNT
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <assert.h>

#ifndef WIN32
#	include <sys/dir.h>
#	include <sys/time.h>
#	include <sys/param.h>       /* DEV_BSIZE */
#	include <values.h>          /* BITSPERBYTE */

#	include <sys/dvh.h>         /* volume header information */
#	include <sys/fs/efs_ino.h>
#	include <sys/fs/efs_fs.h>
#	include <sys/fs/efs_sb.h>
#	include <sys/fs/efs_dir.h>
#	include <unistd.h>
#else
#	include <io.h>
#	include <ctype.h>

#	include <windows.h>

#	include "irix/types.h"
#	include "irix/param.h"
#	define	BITSPERBYTE	8

#	include "irix/dir.h"
#	include "irix/dvh.h"			/* volume header information */
#	include "irix/efs_ino.h"
#	include "irix/efs_fs.h"
#	include "irix/efs_sb.h"
#	include "irix/efs_dir.h"
#endif

#include "irix_efs.h"
#include "irix_io.h"
#include "efs_io.h"


#ifdef WIN32 
DWORD _n_read;
#endif

static int check_dinode(struct efs_dinode *di);
static int scan_block(
	char		*buffer,
	long		osize,
	char		*name,
	efs_ino_t	*inum
	);

int read_extent(
	EFS_FS	*fs,
	extent	*ex,
	char	*buffer
	);

efs_ino_t efs_path_to_inum( const char *ipath )
{
	efs_ino_t			inum;
	efs_ino_t			finum=0;
	EFS_FS				*fs;
	char				*argt;
	OFF_T				start;

	unsigned char		buffer[512 * 256];
	char				path[MAXPATHLEN];

	struct efs_dinode	stdi,
						*di=&stdi;

	long				osize;
	unsigned long		written;

	unsigned int		i;


	strncpy(path, ipath, MAXPATHLEN);

	if ( NULL == (fs = efs_get_default_fs()))
		return 0;

	if ( *path == '/' )
		inum = EFS_ROOTINO;
	else
		inum = fs->inum;

	for(
		argt = strtok( path, "/" );
		argt != NULL;
		argt = strtok( NULL, "/" )
		)
	{
		if ( 0 != efs_read_dinode( fs, inum, buffer, &di ))
			return(0);

		if ( (di->di_mode & IFMT) != IFDIR )
		{
			efs_errno = EFS_ENOTDIR;
			return(0);
		}
		written=0;

		if( di->di_numextents <= EFS_DIRECTEXTENTS )
		{
			for( i = 0; i < di->di_numextents; i++ )
			{
				if ( -1 == ( osize =
					read_extent( fs, &di->di_u.di_extents[i], buffer)))
					break;

				assert( written == 512 * di->di_u.di_extents[i].ex_offset );
				written += (unsigned long) osize;

				if ( 0 != scan_block( buffer, osize, argt, &finum ))
					break; /* error */

				if ( finum != 0 )
					break; /* found! */
			}
		}
		else /* indirect extents */
		{
			for(i = 0; i < di->di_u.di_extents[0].ex_offset; i++)
			{
				int		block;
				int		num_extents_done = 0;

				for( block=0; block<di->di_u.di_extents[i].ex_length; block ++)
				{
					extent	extents[DEV_BSIZE / sizeof(extent)];
					int		ie;

					start = 512 * (
						fs->vh->vh_pt[fs->partition].pt_firstlbn +
						di->di_u.di_extents[i].ex_bn +
						block );

					if ( start != LSEEK(fs->file, start, SEEK_SET))
					{
						efs_errno = EFS_ESEEK;
						break;
					}
                                                      
					if ( DEV_BSIZE != READ( fs->file, extents, DEV_BSIZE ))
					{
						efs_errno = EFS_EREAD;
						break;
					}
#ifdef LITTLEENDIAN
{
					extent *foo = (extent *)extents;
					unsigned int ui;

					while( foo < ((extent *)extents) + di->di_numextents )
					{
						ui=(unsigned int)*((unsigned int *) foo);
						SWAP_WORD(ui);
						foo->ex_magic = ui >> 24;
						foo->ex_bn = ( ui & 0x00FFFFFF );

						ui=(unsigned int)*((unsigned int *) foo + 1);
						SWAP_WORD(ui);
						foo->ex_length = ui >> 24;
						foo->ex_offset = ( ui & 0x00FFFFFF );
						foo++;
					}

}
#endif /* LITTLEENDIAN */
					
					for(
						ie = 0;
						(ie < (DEV_BSIZE / sizeof( extent))) &&
						(num_extents_done < di->di_numextents);
						ie ++, num_extents_done ++
						)
					{
						if ( -1 == ( osize =
						read_extent( fs, &extents[ie], buffer)))
							break;

						assert( written == 512* extents[ie].ex_offset );
						written += (unsigned long) osize;

						if ( 0 != scan_block( buffer, osize, argt, &finum ))
							break; /* error */

						if ( finum != 0 )
							break; /* found! */
					}
					
					if (( finum != 0 )||( efs_errno != EFS_ENOERR ))
						break;
				}
				if (( finum != 0 )||( efs_errno != EFS_ENOERR ))
					break;
			}
		}

		if ( finum == 0 )
		{
			efs_errno = EFS_ENOENT;
			return(0);
		}
		inum = finum;
	}

	return inum;
}

int read_extent(
	EFS_FS	*fs,
	extent	*ex,
	char	*buffer
	)
{
	OFF_T	start;
	off_t	len;

	start = 512 * ( fs->vh->vh_pt[fs->partition].pt_firstlbn + ex->ex_bn );
	len = 512 * ex->ex_length;

	if ( start != LSEEK(fs->file, start, SEEK_SET))
	{
		efs_errno = EFS_ESEEK;
		return(-1);
	}

	if ( len != READ( fs->file, buffer, len ))
	{
		efs_errno = EFS_EREAD;
		return(-1);
	}

	return ( len );
}

int efs_stat(const char *path, struct stat *buf)
{
	unsigned long	inum;
	if ( 0 == (inum = efs_path_to_inum(path)))
		return(-1);

	return( efs_istat( inum, buf ));
}

int efs_istat(efs_ino_t inum, struct stat *buf)
{
	EFS_FS				*fs;

	unsigned char		buffer[DEV_BSIZE];
	struct efs_dinode	*di=NULL;

	if ( NULL == (fs = efs_get_default_fs()))
		return -1;

	if ( 0 != efs_read_dinode( fs, inum, buffer, &di ))
		return -1;

	buf->st_mode = di->di_mode;
#ifdef WIN32
	buf->st_ino = (unsigned short) inum;
#else
	buf->st_ino = inum;
#endif
	buf->st_nlink = di->di_nlink;
	buf->st_uid = di->di_uid;
	buf->st_gid = di->di_gid;
	buf->st_size = di->di_size;
	buf->st_atime = di->di_atime;
	buf->st_mtime = di->di_mtime;
	buf->st_ctime = di->di_ctime;

	return(0);
}

/************************************************************************/

int efs_read_dinode(
	EFS_FS				*fs,
	efs_ino_t			inum,
	unsigned char		buffer[],
	struct efs_dinode	**di
	)
{
	int		cg;
	int		in;
	OFF_T	start;
	OFF_T	remain;

	cg = inum / (fs->efs->fs_cgisize * (512 / sizeof( struct efs_dinode)));
	in = inum % (fs->efs->fs_cgisize * (512 / sizeof( struct efs_dinode)));

	start = 512 * (fs->efs->fs_firstcg + fs->efs->fs_cgfsize * cg)
			+ sizeof( struct efs_dinode) * in +
			fs->vh->vh_pt[fs->partition].pt_firstlbn * 512;

	if ( start % DEV_BSIZE != 0 )
	{
		remain = start % DEV_BSIZE ;
		start -= remain;
	}
	else
	{
		remain = 0L;
	}

	if ( start != LSEEK(fs->file, start, SEEK_SET))
	{
		efs_errno = EFS_ESEEK;
		return(1);
	}

	if ( DEV_BSIZE != READ(fs->file, buffer, DEV_BSIZE ))
	{
		efs_errno = EFS_EREAD;
		return(2);
	}

	if ( *di == NULL )
		*di = (struct efs_dinode *)(buffer + remain);
	else
		memcpy( (char *)*di, buffer + remain, sizeof(struct efs_dinode));

	return (check_dinode(*di));
}


/************************************************************************/

static int check_dinode(struct efs_dinode *di)
{
	int	i;
	int err=0;

#ifdef LITTLEENDIAN
	SWAP_HALF(di->di_mode);
	SWAP_HALF(di->di_nlink);
	SWAP_HALF(di->di_uid);
	SWAP_HALF(di->di_gid);
	SWAP_WORD(di->di_size);
	SWAP_WORD(di->di_atime);
	SWAP_WORD(di->di_mtime);
	SWAP_WORD(di->di_ctime);
	SWAP_WORD(di->di_gen);
	SWAP_HALF(di->di_numextents);

	for ( i = 0; (i<EFS_DIRECTEXTENTS) && (i<di->di_numextents); i++ )
	{
		unsigned int ui;

		ui=(unsigned int)*(unsigned int *) &(di->di_u.di_extents[i]);
		SWAP_WORD(ui);
		di->di_u.di_extents[i].ex_magic = ui >> 24;
		di->di_u.di_extents[i].ex_bn = ( ui & 0x00FFFFFF );

		ui=(unsigned int)*((unsigned int *) &(di->di_u.di_extents[i]) + 1);
		SWAP_WORD(ui);
		di->di_u.di_extents[i].ex_length = ui >> 24;
		di->di_u.di_extents[i].ex_offset = ( ui & 0x00FFFFFF );
	}
#endif /* LITTLEENDIAN */


	for ( i = 0; (i<EFS_DIRECTEXTENTS) && (i<di->di_numextents); i++ )
	{
		if ( 0 != di->di_u.di_extents[i].ex_magic)
		{
			efs_errno = EFS_EMAGIC;
			err++;
		}
	}
	return(err);
}



static int scan_block(
	char		*buffer,
	long		osize,
	char		*name,
	efs_ino_t	*inum )
{
	struct efs_dirblk	*dir;
	int					slot;
	struct efs_dent		*ent;
	int					error=0;


	*inum = 0;
	for(
		dir = (struct efs_dirblk*)buffer;
		dir < (struct efs_dirblk*)(buffer + osize);
		dir ++
		)
	{

#ifdef LITTLEENDIAN
		SWAP_HALF(dir->magic);
#endif /* LITTLEENDIAN */

		if (dir->magic != EFS_DIRBLK_MAGIC)
		{
			efs_errno = EFS_EMAGIC;
			error++;
		}
		else
		{
			for( slot = 0; slot < dir->slots; slot ++)
			{
				if( dir->space[slot] != 0 )
				{
					ent = EFS_SLOTOFF_TO_DEP(
						dir,
						EFS_REALOFF(dir->space[slot])
						);
#ifdef LITTLEENDIAN
					SWAP_HALF(ent->ud_inum.s[0]);
					SWAP_HALF(ent->ud_inum.s[1]);
#endif /* LITTLEENDIAN */

					if (
						( name[ent->d_namelen] == 0x00 ) &&
						( 0 == strncmp( name, ent->d_name, ent->d_namelen ))
						)
					{
/*						*inum = ent->ud_inum.s[1];*/
						*inum = EFS_GET_INUM(ent);
						break;
					}
				}
			}
		}

		if ( *inum != 0 )
			break;
	}
			
	return(error);
}
/************************************************************************/
