/*
 * :set tabstop=4
 * main.c
 * spd@daphne.cps.unizar.es
 * Wed Feb 12 19:57:34 MET 1997
 * Tue Dec  9 00:49:33 MET 1997
 *
 * A program which tries to read EFS by accessing the raw disk device
 * libefs example
 * Tested on IRIX 6.2, and Win NT 4.0 ix86
 * (default)
 *
 * ...or a program which tries to read UFS by accessing the raw disk device
 * libufs example
 * Tested on SunOS 4.1.3_U1, Solaris 2.5 sparc, and Win NT 4.0 ix86
 * (compile with -DUFS)
 *
 * All fs related calculations are macro-based, so you don't need
 * external libraries other than lib[eu]fs
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <errno.h>
#include <time.h>


#ifdef UFS

#	ifdef _SOLARIS_
#		include <sys/time.h>
#		include <unistd.h>
#		include <sys/wait.h>
#		include <sys/param.h>
#		include <sys/dklabel.h>
#		include <sys/fs/ufs_fs.h>
#		include <sys/fs/ufs_inode.h>
#		include <sys/fs/ufs_fsdir.h>

#		include <utime.h>
		extern int optind;

#	endif
#	ifdef _SUNOS_ /* SunOS 4.x */
#		include <sys/time.h>
#		include <unistd.h>
#		include <sys/wait.h>
#		include <sys/param.h>
#		include <sun/dklabel.h>
#		include <ufs/fs.h>
#		include <sys/vnode.h>
#		include <ufs/inode.h>
#		include <ufs/fsdir.h>

#		include <utime.h>
		extern int optind;

#	endif
#	ifdef WIN32
#		include <io.h>
#		include <ctype.h>

#		include <windows.h>

#		include "sunos/s_param.h"
#		include "sunos/s_types.h"
#		include "sunos/s_dklabel.h"
#		include "sunos/s_fs.h"
#		include "sunos/s_vnode.h"
#		include "sunos/s_inode.h"
#		include "sunos/s_fsdir.h"

#		define PATH_SEP        '\\'
#	endif

#	include "sun_ufs.h"
#	include "ufs_db.h"
#	include "ufs_io.h"

#	define	FS_DISK				UFS_DISK
#	define	FS_FS				UFS_FS
#	define	FS_FILE				UFS_FILE
#	define	FS_DIR				UFS_DIR

#	define	FS_FMOUNTSTD		UFS_FMOUNTSTD
#	define	FS_ENOERR			UFS_ENOERR

#	define	fs_direct			direct
#	define	fs_open_disk		ufs_open_disk
#	define	fs_mount			ufs_mount
#	define	fs_getwd			ufs_getwd
#	define	fs_chdir			ufs_chdir
#	define	fs_umount			ufs_umount
#	define	fs_close_disk		ufs_close_disk

#	define	fs_sys_errlist		ufs_sys_errlist
#	define	fs_errno			ufs_errno

#	define	fs_fopen			ufs_fopen
#	define	fs_fread			ufs_fread
#	define	fs_feof				ufs_feof
#	define	fs_fclose			ufs_fclose

#	define	fs_opendir			ufs_opendir
#	define	fs_readdir			ufs_readdir
#	define	fs_istat			ufs_istat
#	define	fs_stat				ufs_stat
#	define	fs_closedir			ufs_closedir

#else /* EFS */
#	ifdef WIN32
#		include <io.h>
#		include <ctype.h>

#		include <windows.h>

#		include "irix/types.h"
#		include "irix/param.h"
#		define	BITSPERBYTE	8

#		include "irix/dir.h"
#		include "irix/dvh.h"			/* volume header information */
#		include "irix/efs_ino.h"
#		include "irix/efs_fs.h"
#		include "irix/efs_sb.h"
#		include "irix/efs_dir.h"

#		define PATH_SEP        '\\'

#	else /* IRIX 6.x */
#		include <sys/dir.h>
#		include <sys/time.h>
#		include <sys/param.h>		/* DEV_BSIZE */
#		include <values.h>			/* BITSPERBYTE */

#		include <sys/dvh.h>			/* volume header information */
#		include <sys/fs/efs_ino.h>
#		include <sys/fs/efs_fs.h>
#		include <sys/fs/efs_sb.h>
#		include <sys/fs/efs_dir.h>
#		include <unistd.h>
#		include <sys/wait.h>
#	endif

#	include <errno.h>
#	include <time.h>

#	include <assert.h>

#	include "irix_efs.h"
#	include "efs_db.h"
#	include "efs_io.h"

#	define	FS_DISK				EFS_DISK
#	define	FS_FS				EFS_FS
#	define	FS_FILE				EFS_FILE
#	define	FS_DIR				EFS_DIR

#	define	FS_FMOUNTSTD		EFS_FMOUNTSTD
#	define	FS_ENOERR			EFS_ENOERR

#	define	fs_direct			efs_direct
#	define	fs_open_disk		efs_open_disk
#	define	fs_mount			efs_mount
#	define	fs_getwd			efs_getwd
#	define	fs_chdir			efs_chdir
#	define	fs_umount			efs_umount
#	define	fs_close_disk		efs_close_disk

#	define	fs_sys_errlist		efs_sys_errlist
#	define	fs_errno			efs_errno

#	define	fs_fopen			efs_fopen
#	define	fs_fread			efs_fread
#	define	fs_feof				efs_feof
#	define	fs_fclose			efs_fclose

#	define	fs_opendir			efs_opendir
#	define	fs_readdir			efs_readdir
#	define	fs_istat			efs_istat
#	define	fs_stat				efs_stat
#	define	fs_closedir			efs_closedir

#endif

#ifdef MY_GETOPT
#	define opterr Opterr
#	define optind Optind
#	define optopt Optopt
#	define optarg Optarg
#	define getopt Getopt
#	define getoptreset Getoptreset

extern int opterr;
extern int optind;
extern int optopt;
extern char *optarg;
extern char getopt( int argc, char  *argv[], char *format );
extern void getoptreset (void);
#endif


#ifdef WIN32
#	include <direct.h>
#	define chdir _chdir
#	define getwd(a) _getcwd(a,sizeof(a))
#	define mkdir(a,b) _mkdir(a)
#endif

#ifndef S_ISDIR
#	define S_ISDIR(mode)   ((mode&S_IFMT) == S_IFDIR)
#endif

char    *app;
int		debug=0;

#ifndef PATH_SEP
#   define PATH_SEP        '/'
#endif


void merror(char *msg);
int cp(int argc, char *argv[]);
void cp_usage(void);
int copy_file(char *in, char *out, int flags);
int copy_dir(char *in, char *out, int flags);
int copy_perm(char *in, char *out);
int ls( char *path, char *args );
int more(char *in);

void usage(void)
{
	fprintf(stderr,"Use: %s [-hd] [-p part] file\n",
	app);
	exit(0);
}


int main(int argc, char ** argv)
{
	extern char         *optarg;
	extern int          optind,opterr;
	int                 opt;

	FS_DISK 			*disk;
	FS_FS 				*fs;

	int                 partition=-1;
	char                *name=NULL;
	char                *path=NULL;
	int					i;
	int					done=0;
	char				cmd_buff[2048];
	char				*cmd;
	int					argtc;
	char				*argt[1024];
	char				**c;
	char				pwd[MAXPATHLEN];

	char				*diskdev;


	if ((app = strrchr(argv[0], PATH_SEP))  != NULL)
		app = app+1;
	else
		app = argv[0];

	while ( (opt=getopt(argc,argv,"hdp:V")) != EOF )
	{
		switch(opt)
		{
			case 'd':
				debug=1;
				break;

			case 'o':
				name=optarg;
				break;

			case 'p':
#ifdef UFS
				if ((*optarg>='a') && (*optarg<='h'))
					partition = (int)(*optarg - 'a');
				else
					partition=atoi(optarg);

				if ((partition<0) || (partition>=NDKMAP))
#else
				partition=atoi(optarg);
				
				if ((partition<0) || (partition>=NPARTAB))
#endif
				{
					fprintf(stderr, "%s: invalid partition (%d)\n",
					app, partition);
				}
				break;
			case 'h':
			default:
				usage();
				break;
		}
	}

	if ( argv[optind] == NULL )
		usage();

/*
 * Aqui comienza el programa de verdaz
 */
	
	diskdev = argv[optind];

	if ( NULL == ( disk = fs_open_disk(diskdev)))
	{
		merror("open disk");
	}
	else
	{
		if (partition == -1)
		{
#ifdef UFS
			fprintf(stderr,"\t%s\n", disk->label->dkl_asciilabel);
			fprintf(stderr,"\t%d Mb\n", ufs_disk_size( disk )/2048);
			for( i=0; i<NDKMAP ; i++)
				if (!ufs_empty_partition( disk, i ))
					fprintf(stderr,"\t%d\t%d Mb\n",
						i, ufs_partition_size( disk, i )/2048);
#else
			/* there is not asciilabel */
			fprintf(stderr,"\t%d Mb\n", efs_disk_size( disk )/2048);
			for( i=0; i<NPARTAB ; i++)
				if ((!efs_empty_partition( disk, i ))
					&&
					(
					(disk->vh->vh_pt[i].pt_type == PTYPE_SYSV)
					||
					(disk->vh->vh_pt[i].pt_type == PTYPE_EFS)
					)
					)
					fprintf(stderr,"\t%d\t%d Mb\n",
						i, efs_partition_size( disk, i )/2048);
#endif
		}
		else
		{
			if ( NULL == ( fs = fs_mount(disk, partition, FS_FMOUNTSTD)))
			{
				merror("mount fs");
			}
			else
			{
#ifdef UFS
				if (debug) ufs_plabel(stderr, fs->label);
				if (debug) ufs_psblock(stderr, fs->ufs);
#else
				if (debug) efs_pvh(stderr, fs->vh);
				if (debug) efs_psblock(stderr, fs->efs);
#endif

				while (!done ) 
				{
					fprintf(stdout, "%s(%d) # ", diskdev, partition);
					fgets(cmd_buff, sizeof(cmd_buff), stdin);
					if (feof(stdin))
					{
						done=1;
						break;
					}
					cmd_buff[strlen(cmd_buff)-1]=0x00;

					if ( cmd_buff[0] == '!' )
						cmd = cmd_buff + 1;
					else
						cmd = cmd_buff;

					argtc=0;
					c = argt;
					for( *c = strtok( cmd, " \t" );
					*c != NULL;
					*(++c) =  strtok(NULL, " \t"))
					{
						argtc++;
					};

					if ( argt[0] == NULL )
						continue;
						

					if ( cmd != cmd_buff )
					{
						int stw;
						int pid;

#ifndef WIN32
						switch ( pid = fork() )
						{
							case -1:
								perror ( "fork" );
								break;
							case 0:
								execvp( argt[0], &argt[0] );
								perror( "execvp" );
								exit(-1);
							default:
								while ( pid != wait ( &stw ) )
								;
								if( stw & 0200 )
									fprintf(stderr,
									"%s: core dumped\n",
									argt[0]);
								break;
						}
#else
						fprintf(stderr,
						"sorry '!' feature not supported on this system\n");
						
#endif
					}
					else if ( !strcmp( argt[0], "ls" ))
					{
						ls(".", argt[1]);
					}
					else if ( !strcmp( argt[0], "bye" ))
					{
						done=1;
					}
					else if ( !strcmp( argt[0], "debug" ))
					{
						if ( debug )
						{
							debug = 0;
							fprintf(stderr,"debug off\n");
						}
						else
						{
							debug = 1;
							fprintf(stderr,"debug on\n");
						}
					}
					else if ( !strcmp( argt[0], "pwd" ))
					{
						fprintf(stderr, "%s\n", fs_getwd(pwd));
					}
					else if ( !strcmp( argt[0], "lcd" ))
					{
						if ( argt[1] == NULL )
						{
							fprintf(stderr, "use: lcd path\n");
						}
						else
						{
							if ( -1 == chdir(argt[1]))
								perror(argt[1]);
						}
					}
					else if ( !strcmp( argt[0], "cd" ))
					{
						if ( argt[1] == NULL )
						{
							fprintf(stderr, "use: cd path\n");
						}
						else
						{
							if ( -1 == fs_chdir( argt[1] ))
							{
								merror(argt[1]);
							}
						}
					}
					else if ( !strcmp( argt[0], "more" ))
					{
						if ( argt[1] == NULL )
						{
							fprintf(stderr, "use: more file\n");
						}
						else
						{
							if ( 0 != more( argt[1] ))
							{
								merror(argt[1]);
							}
						}
					}
					else if ( !strcmp( argt[0], "cp" ))
					{
						if ( 0 != cp( argtc, argt) )
						{
							if ( errno != 0 )
								perror(argt[0]);
						}
					}
					else
					{
						fprintf(stderr,"duh?\n");
						fprintf(stderr,"debug ls bye pwd lcd cd cp more !cmd\n");
					}
				}

				fs_umount( fs );
			}
		}
		fs_close_disk( disk );
	}
	return(0);
}

void merror(char *msg)
{
	if (FS_ENOERR == fs_errno)
		perror(msg);
	else
	{
		fprintf(stderr,"%s: %s\n", msg,
		fs_sys_errlist[fs_errno]);
	}
}

void cp_usage(void)
{
	fprintf(stderr,"Use: cp [-rpv] fs-file unix-file\n");
}

int cp(int argc, char *argv[])
{
#define CP_R	1
#define CP_P	2
#define CP_V	4

	int			opt;

	int			flags=0;

	struct stat	in,out;

	char		cwd[MAXPATHLEN];

	getoptreset();
	while ( (opt=getopt(argc,argv,"rpv")) != EOF )
	{
		switch(opt)
		{
			case 'v':
				if (!( flags & CP_V ))
					flags += CP_V;
				break;
			case 'r':
				if (!( flags & CP_R ))
					flags += CP_R;
				break;
			case 'p':
				if (!( flags & CP_P ))
					flags += CP_P;
				break;
			default:
				cp_usage();
				return(1);
				break;
		}
	}

	if ( argc - optind != 2 )
	{
		fprintf(stderr,"%d:%d\n", optind, argc);
		cp_usage();
		return(2);
	}

	if ( -1 == ( fs_stat(argv[optind], &in)))
	{
		perror(argv[optind]);
		return(3);
	}

	if ( flags & CP_R )
	{
		if (S_ISDIR(in.st_mode))
		{
			if ( -1 == ( stat(argv[optind+1], &out)))
			{
				fprintf(stderr, "%s: directory, target %s does not exist\n",
				argv[optind],argv[optind+1]);
				return(3);
			}

			if ( ! (S_ISDIR(out.st_mode)))
			{
				fprintf(stderr, "%s: directory, target %s not directory\n",
				argv[optind],argv[optind+1]);
				return(4);
			}

			getwd(cwd);

			if ( -1 == chdir( argv[optind+1] ))
			{
				return(5);
			}

			copy_dir( argv[optind], ".", flags );

			chdir(cwd);
		}
		else
		{
			if ( -1 != ( stat(argv[optind+1], &out)))
			{
				if ( S_ISDIR(out.st_mode))
				{
					getwd(cwd);

					if ( -1 == chdir( argv[optind+1] ))
					{
						return(5);
					}

					copy_file( argv[optind], argv[optind], flags );

					chdir(cwd);

				}
				else
				{
					fprintf(stderr, "target %s exists\n", argv[optind+1]);
					return(6);
				}
			}
			fprintf(stderr,"-r makes no sense here\n");
		}
	}
	else
	{
		return( copy_file( argv[optind], argv[optind+1], flags));
	}
	

	return(0);

}

int copy_perm(char *in, char *out)
{
#ifndef WIN32
	struct stat	sbp;
	struct stat	sbo;
	struct utimbuf	times;

	if ( fs_stat(in,&sbp)==-1)
		return(1);
	
	if ( stat(out,&sbo)==-1)
		return(2);
	
	times.actime = sbp.st_atime;
	times.modtime = sbp.st_mtime;
	utime(out, &times);
	chown(out, sbp.st_uid, sbp.st_gid);
	chmod(out, sbp.st_mode);
#endif
	return(0);
}

int copy_dir(char *in, char *out, int flags)
{
	FS_DIR				*dir;
	struct fs_direct	*sd;
	struct stat			st;
	char				pwd[MAXPATHLEN];
	int					err=0;
#ifndef WIN32
	mode_t				mode=0777;
#else
	int					mode=0777;
#endif


	if ( flags & CP_V ) fprintf(stderr,"copy_dir %s %s\n",in, out );
	fs_getwd(pwd);
	if ( -1 == fs_chdir( in ))
	{
		merror(in);
		return(2);
	}

	mkdir(in, mode);
	chdir(in);
	if ( flags & CP_P )
	{
		/* set owner/perm */
		if ( 0 != copy_perm(".", "."))
			perror("copy_perm");
	}

	if ( NULL == (dir = fs_opendir( "." )))
	{
		fprintf(stderr,"opendir %s\n",in);
		merror(in);
		err=1;
		goto quit;
	}


	while ( NULL != ( sd = fs_readdir( dir )))
	{
		if ( -1 == fs_istat( sd->d_ino, &st ))
			merror(sd->d_name);
		else
		{
			switch (st.st_mode & IFMT)
			{
				case IFREG:
					err = copy_file(sd->d_name, sd->d_name, flags);
					break;
				case IFDIR:
					if (!( (sd->d_name[0] == '.') &&
						( (sd->d_name[1] == 0x00) ||
						 ( (sd->d_name[1] == '.') &&
						   (sd->d_name[2] == 0x00)
						   ))))
						err = copy_dir(sd->d_name, ".", flags);
					break;

				case IFIFO:
				case IFCHR:
				case IFBLK:
				case IFLNK:
				case IFSOCK:
				default:
					fprintf(stderr,"%s/%s/%s not copied\n",
						pwd,
						in,
						sd->d_name
						);
			}
		}
	}

quit:
	chdir("..");
	fs_closedir(dir);
	fs_chdir( pwd );
	return err;
}

int copy_file(char *in, char *out, int flags)
{
#define BS	512
	FS_FILE		*i;
	FILE		*o;
	char		buffer[BS];
	size_t		n;

	if ( flags & CP_V ) fprintf(stderr,"copy_file %s %s\n",in, out );

	if ( NULL == ( o = fopen( out, "wb")))
	{
		perror("fopen");
		return 1;
	}

	if ( NULL == ( i = fs_fopen( in, "rb")))
	{
		merror("fs_fopen");
		fclose(o);
		return 1;
	}

	while ( 0 != ( n = fs_fread( buffer, 1, BS, i )))
		fwrite( buffer, 1, n, o );

	
	if (!fs_feof(i))
		merror("fs_fread");

	fs_fclose(i);
	fclose(o);

	if ( flags & CP_P )
	{
		if ( 0 != copy_perm(in, out))
			perror("copy_perm");
	}

	return(0);
}

int more(char *in)
{
	FS_FILE		*i;
	FILE		*o;
	char		buffer[BS];
	size_t		n;

#ifndef WIN32
	if ( NULL == ( o = popen( "less", "wb")))
#else
	if ( NULL == ( o = _popen( "less", "wb")))
#endif
	{
		perror("popen");
		return 1;
	}

	if ( NULL == ( i = fs_fopen( in, "rb")))
	{
		merror("fs_fopen");
		fclose(o);
		return 1;
	}

	while ( 0 != ( n = fs_fread( buffer, 1, BS, i )))
		fwrite( buffer, 1, n, o );

	
	if (!fs_feof(i))
		merror("fs_fread");

	fs_fclose(i);
#ifndef WIN32
	pclose(o);
#else
	_pclose(o);
#endif

	return(0);
}

int ls( char *path, char *args )
{
	FS_DIR 				*dir;
	struct fs_direct	*sd;
	struct stat			st;

	if ( NULL == (dir = fs_opendir( path )))
	{
		fprintf(stderr,"ls %s\n",path);
		merror("opendir");
	}
	else
	{
		while ( NULL != ( sd = fs_readdir( dir )))
		{
			if (args == NULL)
			{
				printf("%10d %24s\n", sd->d_ino, sd->d_name);
			}
			else
			{
				char *ct;

				if ( -1 == fs_istat( sd->d_ino, &st ))
					merror("stat");
				else
				{
					ct = ctime(&st.st_mtime);

					switch (st.st_mode & IFMT)
					{
						case IFIFO:
							fprintf(stderr,"p ");
							break;
						case IFCHR:
							fprintf(stderr,"c ");
							break;
						case IFDIR:
							fprintf(stderr,"d ");
							break;
						case IFBLK:
							fprintf(stderr,"b ");
							break;
						case IFREG:
							fprintf(stderr,"- ");
							break;
						case IFLNK:
							fprintf(stderr,"l ");
							break;
						case IFSOCK:
							fprintf(stderr,"s ");
							break;
						default:
							fprintf(stderr,"? ");
							break;
					}
					fprintf(stderr,"%6d %0.4o %2d %5d %5d %7d %-12.12s %s\n",
					st.st_ino,
					st.st_mode & 0xfff,
					st.st_nlink,
					st.st_uid,
					st.st_gid,
					st.st_size,
					ct + 4,
					sd->d_name);

				}
			}
		}
		fs_closedir( dir );
	}
}
