/* @(#)efs_fs.h Thu Mar 20, 1997 */


#ifdef __cplusplus
extern "C" {
#endif


#define EFS_MAX_DISKS	7
#define EFS_MAX_FSS		16
#define EFS_MAX_DIR		16
#define EFS_MAX_FILE	16

#if defined(__i386__) || defined (M_I86) || defined (_M_IX86)
#	define    i386
#	define    LITTLEENDIAN
#endif

#define swap_word(a) ( ((a) << 24) | \
	(((a) << 8) & 0x00ff0000) | \
	(((a) >> 8) & 0x0000ff00) | \
	((unsigned long)(a) >>24) )

#define swap_half(a) ( ((a & 0xff) << 8) | ((unsigned short)(a) >> 8) )

#define SWAP_WORD(a) { a = swap_word(a); }
#define SWAP_HALF(a) { a = swap_half(a); }

#ifndef MIN
#	define MIN(a,b) ((a)<(b)?(a):(b))
#endif
#define DEV_BSIZE_ROUND (DEV_BSIZE -1)


#ifndef SEEK_SET
#	define SEEK_SET 0
#endif


#ifdef WIN32
#	define	EFS_DISK_T	HANDLE
#else
#	define	EFS_DISK_T	int
#endif

#if defined (__sgi)
#	define efs_direct	direct
#	define efs_dev_t	dev_t
#endif

extern int	efs_errno;

#define EFS_ENOERR		0
#define EFS_ECSUM		1
#define EFS_EMAGIC		2
#define EFS_EREAD		3
#define EFS_EEMPTY		4
#define EFS_EBADPART	5
#define EFS_ESEEK		6
#define EFS_EAGAIN		7
#define EFS_EMFILE		8
#define EFS_ENOTDIR		9
#define EFS_ENAMETOOLONG	10
#define EFS_ENOENT		11

#define EFS_FMOUNTSTD		0
#define EFS_FMOUNTALL		1


extern char *efs_sys_errlist[];

extern struct	_efs_diskbuf
	{
		EFS_DISK_T				file;
		struct volume_header	*vh;
	} _efs_diskb[];

extern struct	_efs_fsbuf
	{
		EFS_DISK_T				file;
		struct volume_header	*vh;
		int						partition;
		struct efs				*efs;

		char					*cwd;
		efs_ino_t				inum;
	} _efs_fsb[];

#define EFS_DISK	struct _efs_diskbuf
#define	EFS_FS		struct _efs_fsbuf


extern struct	_efs_filebuf
	{
		EFS_FS				*fs;
		struct efs_dinode	*di;

		size_t			cnt;   		/* # available characters in buffer */
		unsigned char   *ptr;  		/* next character from here in buffer */
		unsigned char	*base;		/* the buffer */

		unsigned int	i;			/* extent counter */
		int				block;		/* indirect extent block counter */
		int				ie;			/* indirect extent counter */
		int				ed;			/* # extents done */
		extent			*extents;	/* indirect extents */

		size_t			osize;		/* total # characters read */
/*
 * Note:
 * 	Since internal efs_dirblk differs from direct, we need
 *	a static space to hold readdir result. This seems to be
 *	the right way to doit, IRIX 6.x readdir returns always
 *	the same pointer; in SunOS 4.x readdir return different
 *	pointers because the internal format of disk directory
 *	data is the same as struct direct.
 */
		struct efs_direct	dct;		/* for readdir */
		int					slot;		/* slot for efs_dirblk */


	} _efs_fileb[];

		
#define	EFS_DIR		EFS_FILE
#define	EFS_FILE	struct _efs_filebuf
#define	EFS_BLOCK	512

extern EFS_DISK	*efs_open_disk( char *name );
extern int	efs_close_disk( EFS_DISK *disk );

#define efs_empty_partition( disk, i ) \
	((disk)->vh->vh_pt[(i)].pt_nblks ==0 )

#define efs_partition_size( disk, i ) \
	((disk)->vh->vh_pt[(i)].pt_nblks )

#define efs_disk_size( disk ) \
	((disk)->vh->vh_dp.dp_cyls) * 		/* usable cylinders */ \
	((disk)->vh->vh_dp.dp_trks0) * 		/* tracks/cylinder */ \
	((disk)->vh->vh_dp.dp_secs) 		/* sectors/track */

#define efs_disk_size_mbytes( disk ) \
	efs_disk_size( disk ) / 1024  \
	((disk)->vh->vh_dp.dp_secbytes) / 1024	/* bytes/sector */


extern EFS_FS	*efs_mount( EFS_DISK *disk, int partition, int flags );
extern int		efs_umount( EFS_FS *fs );
extern EFS_FS   *efs_get_default_fs( void );
extern void   	efs_set_default_fs( EFS_FS *fs );


/*
 * efs_ ANSI/POSIX like
 */
extern char					*efs_getwd(char *pathname);
extern int					efs_chdir(char *path);

extern EFS_DIR				*efs_opendir( char *name );
extern int					efs_closedir( EFS_DIR *dir );
extern struct efs_direct	*efs_readdir( EFS_DIR *dir );

extern EFS_FILE				*efs_fopen( char *name, char *type );
extern int					efs_fclose( EFS_FILE *stream );
extern size_t				efs_fread(
								void *ptr,
								size_t size,
								size_t nitems,
								EFS_FILE *stream);

extern int efs_stat(const char *path, struct stat *buf);

#define	efs_feof(f)		((f)->osize > (f)->di->di_size)

#ifdef __cplusplus
}
#endif

