/*
 * :set tabstop=4
 * main.c
 * spd@daphne.cps.unizar.es
 * Wed Feb 12 19:57:34 MET 1997
 *
 * A program which tries to read EFS by accessing the raw disk device
 * libefs example
 * Tested on IRIX 6.2, and Win NT 4.0 ix86
 *
 * All fs related calculations are macro-based, so you don't need
 * external libraries
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/dir.h>

#include <errno.h>
#include <time.h>

#ifdef WIN32
#	include <io.h>
#	include <ctype.h>

#	include <windows.h>

extern int optind;
extern char getopt( int argc, char  *argv[], char *format );

#	include "irix/types.h"
#	include "irix/param.h"
#	define	BITSPERBYTE	8

#	include "irix/dvh.h"			/* volume header information */
#	include "irix/efs_ino.h"
#	include "irix/efs_fs.h"
#	include "irix/efs_sb.h"
#	include "irix/efs_dir.h"

#   define PATH_SEP        '\\'
#else
#	include <sys/time.h>
#	include <sys/param.h>		/* DEV_BSIZE */
#	include <values.h>			/* BITSPERBYTE */

#	include <sys/dvh.h>			/* volume header information */
#	include <sys/fs/efs_ino.h>
#	include <sys/fs/efs_fs.h>
#	include <sys/fs/efs_sb.h>
#	include <sys/fs/efs_dir.h>
#	include <unistd.h>
#endif

#include <errno.h>
#include <time.h>

#include <assert.h>

#include "irix_efs.h"
#include "efs_db.h"

char    *app;
int		debug=0;

#ifndef PATH_SEP
#   define PATH_SEP        '/'
#endif


void merror(char *msg);
int copy(char *in, char *out);

void usage(void)
{
	fprintf(stderr,"Use: %s [-hd] [-p part] [-i inode] [-of file] file\n",
	app);
	exit(0);
}


int main(int argc, char ** argv)
{
	extern char         *optarg;
	extern int          optind,opterr;
	int                 opt;

	EFS_DISK 			*disk;
	EFS_FS 				*fs;
	EFS_DIR 			*dir;
	struct efs_direct	*sd;

	int                 partition=-1;
	char                *name=NULL;
	char                *path=NULL;

	unsigned long       inum = EFS_ROOTINO;
	int					i;

	struct stat			st;


	if ((app = strrchr(argv[0], PATH_SEP))  != NULL)
		app = app+1;
	else
		app = argv[0];

	while ( (opt=getopt(argc,argv,"hrdp:i:f:Vo:")) != EOF )
	{
		switch(opt)
		{
			case 'd':
				debug=1;
				break;

			case 'f':
				path=optarg;
				break;

			case 'o':
				name=optarg;
				break;

			case 'i':
				inum=(unsigned long) atol(optarg);
				if (inum<0)
				{
					fprintf(stderr, "%s: invalid inode (%d)\n",
					app, inum);
					exit(1);
				}
				break;
			case 'p':
				partition=atoi(optarg);
				
				if ((partition<0) || (partition>=NPARTAB))
				{
					fprintf(stderr, "%s: invalid partition (%d)\n",
					app, partition);
				}
				break;
			case 'h':
			default:
				usage();
				break;
		}
	}

	if ( argv[optind] == NULL )
		usage();

/*
 * Aqui comienza el programa de verdaz
 */
	
	if ( NULL == ( disk = efs_open_disk(argv[optind])))
	{
		merror("open disk");
	}
	else
	{
		if (partition == -1)
		{
			/* there is not asciilabel */
			fprintf(stderr,"\t%d Mb\n", efs_disk_size( disk )/2048);
			for( i=0; i<NPARTAB ; i++)
				if ((!efs_empty_partition( disk, i ))
					&&
					(
					(disk->vh->vh_pt[i].pt_type == PTYPE_SYSV)
					||
					(disk->vh->vh_pt[i].pt_type == PTYPE_EFS)
					)
					)
					fprintf(stderr,"\t%d\t%d Mb\n",
						i, efs_partition_size( disk, i )/2048);
		}
		else
		{
			if ( NULL == ( fs = efs_mount(disk, partition, EFS_FMOUNTSTD)))
			{
				merror("efs_mount");
			}
			else
			{
				char                            pwd[MAXPATHLEN];

				if (debug) efs_pvh(stderr, fs->vh);
				if (debug) efs_psblock(stderr, fs->efs);

				printf("getwd: %s\n", efs_getwd(pwd));

				if (path)
				{
				printf("chdir: %s\n", path);
				if ( -1 == efs_chdir( path ))
				{
					merror(path);
				}
				printf("getwd: %s\n", efs_getwd(pwd));

				if ( -1 == efs_stat( path, &st ))
					merror("stat");
				else
					if (debug) efs_pstat(stdout, &st);

				
				puts(path);
				if ( NULL == (dir = efs_opendir( path )))
/*				if ( NULL == (dir = efs_opendir( "." )))*/
				{
					puts(path);
					merror("opendir");
				}
				else
				{
					while ( NULL != ( sd = efs_readdir( dir )))
					{
						printf("DIR: %10d %24s\n", sd->d_ino, sd->d_name);
					}
					efs_closedir( dir );
				}
				}
				printf("XXX\n");

				if(name)
					copy( name, name );

				efs_umount( fs );
			}
		}
		efs_close_disk( disk );
	}
	return(0);
}

void merror(char *msg)
{
	if (EFS_ENOERR == efs_errno)
		perror(msg);
	else
	{
		fprintf(stderr,"%s: %s\n", msg,
		efs_sys_errlist[efs_errno]);
	}
}

int copy(char *in, char *out)
{
#define BS	512


	EFS_FILE	*i;
	FILE		*o;
	char		buffer[BS];
	size_t		n;

	if ( NULL == ( o = fopen( out, "wb")))
	{
		perror("fopen");
		return 1;
	}

	if ( NULL == ( i = efs_fopen( in, "rb")))
	{
		merror("efs_fopen");
		fclose(o);
		return 1;
	}
	if (debug) {fprintf(stdout, "OPEN: %s\n",in); fflush(stdout);}

	while ( 0 != ( n = efs_fread( buffer, 1, BS, i )))
	{
		if (debug) fprintf(stdout, "IO: %d\n",n);
		fwrite( buffer, 1, n, o );
	}
	
	if (!efs_feof(i))
		merror("efs_fread");

	efs_fclose(i);
	fclose(o);

	return(0);
}
