/* @(#)ufs_fs.h Thu Mar 20, 1997 */


#ifdef __cplusplus
extern "C" {
#endif


#define UFS_MAX_DISKS	7
#define UFS_MAX_FSS		16
#define UFS_MAX_DIR		16
#define UFS_MAX_FILE	16

#if defined(__i386__) || defined (M_I86) || defined (_M_IX86)
#	define    i386
#	define    LITTLEENDIAN
#endif

#define swap_word(a) ( ((a) << 24) | \
	(((a) << 8) & 0x00ff0000) | \
	(((a) >> 8) & 0x0000ff00) | \
	((unsigned long)(a) >>24) )

#define swap_half(a) ( ((a & 0xff) << 8) | ((unsigned short)(a) >> 8) )

#define SWAP_WORD(a) { a = swap_word(a); }
#define SWAP_HALF(a) { a = swap_half(a); }

#ifndef MIN
#	define MIN(a,b) ((a)<(b)?(a):(b))
#endif
#define DEV_BSIZE_ROUND (DEV_BSIZE -1)


#ifndef SEEK_SET
#	define SEEK_SET 0
#endif


#ifdef WIN32
#	define	UFS_DISK_T	HANDLE
#else
#	define	UFS_DISK_T	int
#endif

extern int	ufs_errno;

#define UFS_ENOERR		0
#define UFS_ECSUM		1
#define UFS_EMAGIC		2
#define UFS_EREAD		3
#define UFS_EEMPTY		4
#define UFS_EBADPART	5
#define UFS_ESEEK		6
#define UFS_EAGAIN		7
#define UFS_EMFILE		8
#define UFS_ENOTDIR		9
#define UFS_ENAMETOOLONG	10
#define UFS_ENOENT		11

#define UFS_FMOUNTSTD		0
#define UFS_FMOUNTALL		1


extern char *ufs_sys_errlist[];

extern struct	_ufs_diskbuf
	{
		UFS_DISK_T		file;
		struct dk_label	*label;
	} _ufs_diskb[];

extern struct	_ufs_fsbuf
	{
		UFS_DISK_T		file;
		struct dk_label	*label;
		int				partition;
		struct fs		*ufs;

		char			*cwd;
		unsigned long	inum;
	} _ufs_fsb[];

#define UFS_DISK	struct _ufs_diskbuf
#define	UFS_FS		struct _ufs_fsbuf


extern struct	_ufs_filebuf
	{
		UFS_FS			*fs;
		struct dinode	*di;
		size_t			cnt;   /* number of available characters in buffer */
		unsigned char   *ptr;  /* next character from here in buffer */
		unsigned char	*base; /* the buffer */
		unsigned char	*fsb;  /* indirect block */
		daddr_t			*ibno; /* indirect block inode */
		long			osize; /* bytes read */
		int				lbn;   /* block number */
		int				ilbn;  /* indirect block counter */	
	} _ufs_fileb[];

		
#define	UFS_DIR		UFS_FILE
#define	UFS_FILE	struct _ufs_filebuf
#define	UFS_BLOCK	512

extern UFS_DISK	*ufs_open_disk( char *name );
extern int	ufs_close_disk( UFS_DISK *disk );

#define ufs_empty_partition( disk, i ) \
	((disk)->label->dkl_map[(i)].dkl_nblk ==0 )

#define ufs_partition_size( disk, i ) \
	((disk)->label->dkl_map[(i)].dkl_nblk)

#define ufs_disk_size( disk ) \
	((disk)->label->dkl_ncyl * \
	(disk)->label->dkl_nhead * \
	(disk)->label->dkl_nsect )

extern UFS_FS	*ufs_mount( UFS_DISK *disk, int partition, int flags );
extern int		ufs_umount( UFS_FS *fs );
extern UFS_FS   *get_default_fs( void );
extern void   	set_default_fs( UFS_FS *fs );


/*
 * ufs_ ANSI/POSIX like
 */
extern char		*ufs_getwd(char *pathname);
extern int		ufs_chdir(char *path);

extern UFS_DIR			*ufs_opendir( char *name );
extern int				ufs_closedir( UFS_DIR *dir );
extern struct direct	*ufs_readdir( UFS_DIR *dir );

extern UFS_FILE		*ufs_fopen( char *name, char *type );
extern int			ufs_fclose( UFS_FILE *stream );
extern size_t		ufs_fread(
						void *ptr,
						size_t size,
						size_t nitems,
						UFS_FILE *stream);

extern int ufs_stat(const char *path, struct stat *buf);

#define	ufs_feof(f)		(( f->ilbn == NIADDR ) && (f->cnt==0))

#ifdef __cplusplus
}
#endif

