/*
 * :set tabstop=4
 * ufs.c
 * spd@daphne.cps.unizar.es
 * Wed Feb 12 19:57:34 MET 1997
 * Mon Dec  8 23:47:34 MET 1997
 * Sun Sep 13 13:22:02 MET DST 1998 (lld, level1 & 2)
 *
 * A program which tries to read UFS by accessing the raw disk device
 * Version for WinNT, SunOS 4.x and Solaris 2.x
 *
 * All fs related calculations are macro-based, so you don't need
 * external libraries
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <errno.h>
#include <time.h>

#if defined(__i386__) || defined (M_I86) || defined (_M_IX86)
# define    i386
# define    LITTLEENDIAN
#endif

#	ifdef _SOLARIS_
#		include <sys/time.h>
#		include <unistd.h>
#		include <sys/wait.h>
#		include <sys/param.h>
#		include <sys/dklabel.h>
#		include <sys/fs/ufs_fs.h>
#		include <sys/fs/ufs_inode.h>
#		include <sys/fs/ufs_fsdir.h>
#	endif
#	ifdef _SUNOS_ /* SunOS 4.x */
#		include <sys/time.h>
#		include <unistd.h>
#		include <sys/wait.h>
#		include <sys/param.h>
#		include <sun/dklabel.h>
#		include <ufs/fs.h>
#		include <sys/vnode.h>
#		include <ufs/inode.h>
#		include <ufs/fsdir.h>
#	endif
#	ifdef WIN32
#		include <io.h>
#		include <ctype.h>
#		include <assert.h>

#		include <windows.h>

#		include "s_param.h"
#		include "s_types.h"
#		include "s_dklabel.h"
#		include "s_fs.h"
#		include "s_vnode.h"
#		include "s_inode.h"
#		include "s_fsdir.h"

		extern int optind;
		extern char getopt( int argc, char  *argv[], char *format );
#		define PATH_SEP        '\\'

#	endif

#ifndef WIN32
#	define  DISK_T  int
#	define  OFF_T   off_t
#	define  LSEEK(f, o, w)  lseek(f, o, w)
#	define  READ(file, buff, nbyte) read(file, buff, nbyte)
	#define  CLOSE close
#else
DWORD       _n_read;

#	define  DISK_T  HANDLE
#	define  OFF_T   off_t
#	define  LSEEK(f, o, w) \
	((SetFilePointer(f, o, NULL, FILE_BEGIN) == 0xFFFFFFFF) ? \
	-1 : o )
#	define  READ(file, buff, nbyte) \
	(ReadFile( file, buff, nbyte, (LPDWORD)&_n_read, NULL ) ? \
	_n_read : -1 )
#	define  CLOSE CloseHandle
#endif


static char __ident[] = "@(#)(c)SPDsoft, Mon Dec 8, 1997";
#define VERS_STR        ((char*)&__ident[4])

#ifdef _BSD_
	extern char *sys_errlist[];
#	define strerror(e) sys_errlist[(e)]
#endif
#ifndef PATH_SEP
#	define PATH_SEP        '/'
#endif
#ifndef SEEK_SET
#	define SEEK_SET 0
#endif

#define	fsdbtoaddr(a,label,fs,partition)	\
	((fsbtodb((fs),(a)) + \
	(label)->dkl_map[(partition)].dkl_cylno * \
	(label)->dkl_nhead * \
	(label)->dkl_nsect ) * 512)

#define swap_word(a) ( ((a) << 24) | \
    (((a) << 8) & 0x00ff0000) | \
    (((a) >> 8) & 0x0000ff00) | \
    ((unsigned long)(a) >>24) )

#define swap_half(a) ( ((a & 0xff) << 8) | ((unsigned short)(a) >> 8) )

#define SWAP_WORD(a) { a = swap_word(a); }
#define SWAP_HALF(a) { a = swap_half(a); }

#ifndef MIN
#	define MIN(a,b) ((a)<(b)?(a):(b))
#endif
#define DEV_BSIZE_ROUND	(DEV_BSIZE -1)

/*extern int optind;*/
/*extern char getopt( int argc, char  *argv[], char *format );*/

/*
 * Prototypes (start)
 */
void merror(char *msg);
void usage(void);
int		check_dk_label(struct dk_label *label);
int		check_sblock(struct fs *ufs);
int		check_dinode(struct dinode *di);

int read_dinode(
	DISK_T				disk,
	struct dk_label	*label,
	int				partition,
	struct fs		*ufs,
	unsigned long	inum,
	unsigned char	buffer[],

	struct dinode	**di
	);

int read_fsblock(
	DISK_T				disk,
	struct dk_label	*label,
	int				partition,
	struct fs		*ufs,

	struct dinode	*di,

	daddr_t			block,
	int				f,
	unsigned char	*fsdb,
	long			*osize,
	int				*lbn
	);

/*
 * Prototypes (start)
 */

/*
 * Globals
 */
extern int errno;
char    *app;
int		debug=0;
int		raw=0;


/*
 * Fun stuff
 */
void merror(char *msg)
{
#ifdef WIN32
 	char * lpMsgBuf;

	FormatMessage(
			FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
			NULL,
			GetLastError(),
			MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
			(LPTSTR) &lpMsgBuf,
			0,
			NULL
		);

	/* Display the string.	*/
/*	MessageBox( NULL, lpMsgBuf, msg, MB_OK|MB_ICONINFORMATION );*/
	fprintf( stderr, "%s: %s", msg, lpMsgBuf );

	LocalFree( lpMsgBuf );
#else
	perror(msg);
#endif
}

void usage(void)
{
	fprintf(stderr,"Use: %s [-hrd] [-p partition] [-i inode] [-o file] file\n",
	app);
	exit(0);
}

int		check_dinode(struct dinode *di)
{

	int i;
#ifdef LITTLEENDIAN

	SWAP_HALF(di->di_mode);

	SWAP_HALF(di->di_nlink);
	SWAP_HALF(di->di_uid);
	SWAP_HALF(di->di_gid);
	SWAP_WORD(di->di_size);

	SWAP_WORD(di->di_atime);
	SWAP_WORD(di->di_ic.ic_atspare);
	SWAP_WORD(di->di_mtime);
	SWAP_WORD(di->di_ic.ic_mtspare);
	SWAP_WORD(di->di_ctime);
	SWAP_WORD(di->di_ic.ic_ctspare);

	for(i=0; i<NDADDR; i++)
		SWAP_WORD(di->di_db[i]);

	for(i=0; i<NIADDR; i++)
		SWAP_WORD(di->di_ib[i]);

	SWAP_WORD(di->di_ic.ic_flags);
	SWAP_WORD(di->di_blocks);
	SWAP_WORD(di->di_gen);

#endif /* LITTLEENDIAN */

	if (debug)
	{

fprintf(stderr,"%x\tmode and type of file\n", di->di_mode);

switch (di->di_mode & IFMT)
{
	case IFIFO:
		fprintf(stderr,"\t\tnamed pipe (fifo)\n");
		break;
	case IFCHR:
		fprintf(stderr,"\t\tcharacter special\n");
		break;
	case IFDIR:
		fprintf(stderr,"\t\tdirectory\n");
		break;
	case IFBLK:
		fprintf(stderr,"\t\tblock special\n");
		break;
	case IFREG:
		fprintf(stderr,"\t\tregular\n");
		break;
	case IFLNK:
		fprintf(stderr,"\t\tsymbolic link\n");
		break;
	case IFSOCK:
		fprintf(stderr,"\t\tsocket\n");
		break;
	default:
		fprintf(stderr,"\t\tunknown\n");
		break;
}
		
if (di->di_mode & ISUID)
	fprintf(stderr,"\t\tset user id on execution\n");
if (di->di_mode & ISGID)
	fprintf(stderr,"\t\tset group id on execution\n");
if (di->di_mode & ISVTX)
	fprintf(stderr,"\t\tsave swapped text even after use\n");
fprintf(stderr,"\t\tPermissions: %o\n", di->di_mode & 0x0fff);

fprintf(stderr,"%d\tnumber of links to file\n", di->di_nlink);
fprintf(stderr,"%d\towner's user id\n", di->di_uid);
fprintf(stderr,"%d\towner's group id\n", di->di_gid);
fprintf(stderr,"%lld\tnumber of bytes in file\n", di->di_size);

fprintf(stderr,"\ttime last accessed\t%s", ctime((time_t *)&di->di_atime));
fprintf(stderr,"\tatspare\t%s", ctime((time_t *)&di->di_ic.ic_atspare));
fprintf(stderr,"\ttime last modified\t%s", ctime((time_t *)&di->di_mtime));
fprintf(stderr,"\tmtspare\t%s", ctime((time_t *)&di->di_ic.ic_mtspare));
fprintf(stderr,"\tlast time inode changed\t%s", ctime((time_t *)&di->di_ctime));
fprintf(stderr,"\tctspare\t%s", ctime((time_t *)&di->di_ic.ic_ctspare));
fprintf(stderr,"\t\tic_db[0]=di_rdev\n");

	for(i=0; i<NDADDR; i++)
		fprintf(stderr,"\t\tdisk block address ic_db[%d]=0x%x\n", i,
		di->di_db[i]
		);
	for(i=0; i<NIADDR; i++)
		fprintf(stderr,"\t\tindirect block ic_ib[%d]=%x\n", i,
		di->di_ib[i]
		);

fprintf(stderr,"%d\tstatus, currently unused\n", di->di_ic.ic_flags);
fprintf(stderr,"%d\tblocks actually held\n", di->di_blocks);
fprintf(stderr,"%d\tgeneration number\n", di->di_gen);
		/*
		 * XXX - the disk spares were used to avoid changing
		 * the size of the incore inode in a minor release.
		 * Fix for 5.0 release.
		 * Also remove the code in iget, iupdat that clears them.
		 */
#ifdef _SOLARIS_
fprintf(stderr,"%d\tshadow inode\n", di->di_ic.ic_shadow);
fprintf(stderr,"%d\tlong EFT version of uid\n", di->di_uid);
fprintf(stderr,"%d\tlong EFT version of gid\n", di->di_gid);
fprintf(stderr,"%d\treserved (ic_oeftflag)\n", di->di_ic.ic_oeftflag);
#else
fprintf(stderr,"%d\tdelayed writes, units=bytes\n", di->di_ic.ic_delaylen);
fprintf(stderr,"%d\twhere we started delaying\n", di->di_ic.ic_delayoff);
fprintf(stderr,"%d\twhere to start the next clust\n", di->di_ic.ic_nextrio);
fprintf(stderr,"%d\tnumber of outstanding writes\n", di->di_ic.ic_writes);
#endif
	}
	return(0);
}

int check_dk_label(struct dk_label *label)
{
	short				i;
	unsigned short		sum, *word;

	for (
		word = (unsigned short *)label, sum=0;
		word < ((unsigned short *)label + sizeof(*label)/sizeof(short) -1);
		word ++
		)
		sum ^= *word;

#ifdef LITTLEENDIAN

	SWAP_HALF(label->dkl_rpm);
	SWAP_HALF(label->dkl_pcyl);
	SWAP_HALF(label->dkl_apc);
	SWAP_HALF(label->dkl_intrlv);
	SWAP_HALF(label->dkl_ncyl);
	SWAP_HALF(label->dkl_acyl);
	SWAP_HALF(label->dkl_nhead);
	SWAP_HALF(label->dkl_nsect);
	for( i=0; i<NDKMAP ; i++)
	{
		SWAP_WORD(label->dkl_map[i].dkl_cylno);
		SWAP_WORD(label->dkl_map[i].dkl_nblk);
	}
	SWAP_HALF(label->dkl_magic);

#endif /* LITTLEENDIAN */

	if (debug)
	{
	fprintf(stderr,"%s\n",label->dkl_asciilabel);
	/* dkl_pad */
	fprintf(stderr,"%u\trotations per minute\n",label->dkl_rpm);
	fprintf(stderr,"%u\tphysical cylinders\n",label->dkl_pcyl);
	fprintf(stderr,"%u\talternates per cylinder\n",label->dkl_apc);
	fprintf(stderr,"%x\tobsolete\n",label->dkl_obs1);
	fprintf(stderr,"%x\tobsolete\n",label->dkl_obs2);
	fprintf(stderr,"%u\tinterleave factor\n",label->dkl_intrlv);
	fprintf(stderr,"%u\tdata cylinders\n",label->dkl_ncyl);
	fprintf(stderr,"%u\talternate cylinders\n",label->dkl_acyl);
	fprintf(stderr,"%u\theads in this partition\n",label->dkl_nhead);
	fprintf(stderr,"%u\t512 byte sectors per track",label->dkl_nsect);
	fprintf(stderr,"\t(This makes a %d Mb disk)\n",
		label->dkl_ncyl * label->dkl_nhead *
		label->dkl_nsect * 512 / 1024 / 1024
		);
	fprintf(stderr,"%x\tobsolete\n",label->dkl_obs3);
	fprintf(stderr,"%x\tobsolete\n",label->dkl_obs4);
	for( i=0; i<NDKMAP ; i++)
	{
		if ( (label->dkl_map[i].dkl_nblk !=0 ) ||
			(label->dkl_map[i].dkl_cylno != 0) )
		{
			fprintf(stderr,"\t%d\tlogical partition\n", i);
			fprintf(stderr,"\t\t%d\tstarting cylinder (0x%x)\n",
				label->dkl_map[i].dkl_cylno,
				label->dkl_map[i].dkl_cylno *
				label->dkl_nhead *
				label->dkl_nsect * 512
				);
			fprintf(stderr,"\t\t%d\tblocks (%d b, %d Kb, %d Mb)\n",
				label->dkl_map[i].dkl_nblk,
				label->dkl_map[i].dkl_nblk*512,
				label->dkl_map[i].dkl_nblk*512/1024,
				label->dkl_map[i].dkl_nblk*512/1024/1024
				);
		}
	}
	fprintf(stderr,"0x%X\tmagic (should be 0x%X)\n",
		label->dkl_magic, DKL_MAGIC);

	fprintf(stderr,"0x%X\txor checksum of sector (calculated 0x%X)\n",
	label->dkl_cksum, sum);
	}

	return (( sum == label->dkl_cksum ) && ( DKL_MAGIC == label->dkl_magic ));
}

int		check_sblock(struct fs *ufs)
{
#ifdef LITTLEENDIAN
	SWAP_WORD(ufs->fs_sblkno);
	SWAP_WORD(ufs->fs_cblkno);
	SWAP_WORD(ufs->fs_iblkno);
	SWAP_WORD(ufs->fs_dblkno);
	SWAP_WORD(ufs->fs_cgoffset);
	SWAP_WORD(ufs->fs_cgmask);

	SWAP_WORD(ufs->fs_time);

	SWAP_WORD(ufs->fs_size);
	SWAP_WORD(ufs->fs_dsize);

	SWAP_WORD(ufs->fs_ncg);
	SWAP_WORD(ufs->fs_bsize);
	SWAP_WORD(ufs->fs_fsize);
	SWAP_WORD(ufs->fs_frag);
	SWAP_WORD(ufs->fs_minfree);
	SWAP_WORD(ufs->fs_rotdelay);
	SWAP_WORD(ufs->fs_rps);
	SWAP_WORD(ufs->fs_bmask);
	SWAP_WORD(ufs->fs_fmask);
	SWAP_WORD(ufs->fs_bshift);
	SWAP_WORD(ufs->fs_fshift);
	SWAP_WORD(ufs->fs_maxcontig);
	SWAP_WORD(ufs->fs_maxbpg);
	SWAP_WORD(ufs->fs_fragshift);
	SWAP_WORD(ufs->fs_fsbtodb);
	SWAP_WORD(ufs->fs_sbsize);
	SWAP_WORD(ufs->fs_csmask);
	SWAP_WORD(ufs->fs_csshift);
	SWAP_WORD(ufs->fs_nindir);
	SWAP_WORD(ufs->fs_inopb);
	SWAP_WORD(ufs->fs_nspf);
	SWAP_WORD(ufs->fs_optim);
	SWAP_WORD(ufs->fs_npsect);
	SWAP_WORD(ufs->fs_interleave);
	SWAP_WORD(ufs->fs_trackskew);
	SWAP_WORD(ufs->fs_id[0]);
	SWAP_WORD(ufs->fs_id[1]);
	SWAP_WORD(ufs->fs_csaddr);
	SWAP_WORD(ufs->fs_cssize);
	SWAP_WORD(ufs->fs_cgsize);
	SWAP_WORD(ufs->fs_ntrak);
	SWAP_WORD(ufs->fs_nsect);
	SWAP_WORD(ufs->fs_spc);
	SWAP_WORD(ufs->fs_ncyl);
	SWAP_WORD(ufs->fs_cpg);
	SWAP_WORD(ufs->fs_ipg);

	SWAP_WORD(ufs->fs_fpg);

	SWAP_WORD(ufs->fs_cstotal.cs_ndir);
	SWAP_WORD(ufs->fs_cstotal.cs_nbfree);
	SWAP_WORD(ufs->fs_cstotal.cs_nifree);
	SWAP_WORD(ufs->fs_cstotal.cs_nffree);

	SWAP_WORD(ufs->fs_cgrotor);

	SWAP_WORD(ufs->fs_cpc);

	SWAP_WORD(ufs->fs_qbmask.val[0]);
	SWAP_WORD(ufs->fs_qbmask.val[1]);
	SWAP_WORD(ufs->fs_qfmask.val[0]);
	SWAP_WORD(ufs->fs_qfmask.val[1]);
	SWAP_WORD(ufs->fs_postblformat);
	SWAP_WORD(ufs->fs_nrpos);
	SWAP_WORD(ufs->fs_postbloff);
	SWAP_WORD(ufs->fs_rotbloff);
	SWAP_WORD(ufs->fs_magic);

#endif /* LITTLEENDIAN */

	if (debug)
	{
fprintf(stderr,"%x\tlinked list of file systems\n",ufs->fs_link);
fprintf(stderr,"%x\tused for incore super blocks\n",ufs->fs_rlink);
fprintf(stderr,"%x\taddr of super-block in filesys\n",ufs->fs_sblkno);
fprintf(stderr,"%x\toffset of cyl-block in filesys\n",ufs->fs_cblkno);
fprintf(stderr,"%x\toffset of inode-blocks in filesys\n",ufs->fs_iblkno);
fprintf(stderr,"%x\toffset of first data after cg\n",ufs->fs_dblkno);
fprintf(stderr,"%x\tcylinder group offset in cylinder\n",ufs->fs_cgoffset);
fprintf(stderr,"%x\tused to calc mod fs_ntrak\n",ufs->fs_cgmask);
fprintf(stderr,"last time written:\t%s",ctime(&ufs->fs_time));
fprintf(stderr,"%u\tnumber of blocks in fs (%d b, %d Kb, %d Mb)\n",
			ufs->fs_size,
			ufs->fs_size*fsbtodb(ufs, 1)*512,
			ufs->fs_size*fsbtodb(ufs, 1)*512/1024,
			ufs->fs_size*fsbtodb(ufs, 1)*512/1024/1024
			);
fprintf(stderr,"%u\tnumber of data blocks in fs (%d b, %d Kb, %d Mb) \n",
			ufs->fs_dsize,
			ufs->fs_dsize*fsbtodb(ufs, 1)*512,
			ufs->fs_dsize*fsbtodb(ufs, 1)*512/1024,
			ufs->fs_dsize*fsbtodb(ufs, 1)*512/1024/1024
			);


fprintf(stderr,"%u\tnumber of cylinder groups\n",ufs->fs_ncg);
fprintf(stderr,"%u\tsize of basic blocks in fs\n",ufs->fs_bsize);
fprintf(stderr,"%u\tsize of frag blocks in fs\n",ufs->fs_fsize);
fprintf(stderr,"%u\tnumber of frags in a block in fs\n",ufs->fs_frag);
fprintf(stderr,"%u\tminimum percentage of free blocks\n",ufs->fs_minfree);
fprintf(stderr,"%u\tnum of ms for optimal next block\n",ufs->fs_rotdelay);
fprintf(stderr,"%u\tdisk revolutions per second\n",ufs->fs_rps);
fprintf(stderr,"%x\t``blkoff'' calc of blk offsets\n",ufs->fs_bmask);
fprintf(stderr,"%x\t``fragoff'' calc of frag offsets\n",ufs->fs_fmask);
fprintf(stderr,"%x\t``lblkno'' calc of logical blkno\n",ufs->fs_bshift);
fprintf(stderr,"%x\t``numfrags'' calc number of frags\n",ufs->fs_fshift);
fprintf(stderr,"%u\tmax number of contiguous blks\n",ufs->fs_maxcontig);
fprintf(stderr,"%u\tmax number of blks per cyl group\n",ufs->fs_maxbpg);
fprintf(stderr,"%u\tblock to frag shift\n",ufs->fs_fragshift);
fprintf(stderr,"%u\tfsbtodb and dbtofsb shift constant\n",ufs->fs_fsbtodb);
fprintf(stderr,"%u\tactual size of super block\n",ufs->fs_sbsize);
fprintf(stderr,"%x\tcsum block offset\n",ufs->fs_csmask);
fprintf(stderr,"%x\tcsum block number\n",ufs->fs_csshift);
fprintf(stderr,"%u\tvalue of NINDIR\n",ufs->fs_nindir);
fprintf(stderr,"%u\tvalue of INOPB\n",ufs->fs_inopb);
fprintf(stderr,"%u\tvalue of NSPF\n",ufs->fs_nspf);
fprintf(stderr,"%x\toptimization preference, see below\n",ufs->fs_optim);
fprintf(stderr,"%u\t# sectors\n",ufs->fs_npsect);
fprintf(stderr,"%u\thardware sector interleave\n",ufs->fs_interleave);
fprintf(stderr,"%u\tsector 0 skew, per track\n",ufs->fs_trackskew);
fprintf(stderr,"%x %x\tfile system id\n",ufs->fs_id[0],ufs->fs_id[1]);
fprintf(stderr,"%x\tblk addr of cyl grp summary area\n",ufs->fs_csaddr);
fprintf(stderr,"%u\tsize of cyl grp summary area\n",ufs->fs_cssize);
fprintf(stderr,"%u\tcylinder group size\n",ufs->fs_cgsize);
fprintf(stderr,"%u\ttracks per cylinder\n",ufs->fs_ntrak);
fprintf(stderr,"%u\tsectors per track\n",ufs->fs_nsect);
fprintf(stderr,"%u\tsectors per cylinder (%d*%d=%d)\n",
		ufs->fs_spc,
		ufs->fs_ntrak,ufs->fs_nsect,ufs->fs_ntrak*ufs->fs_nsect
		);
fprintf(stderr,"%u\tcylinders in file system\n",ufs->fs_ncyl);
fprintf(stderr,"%u\tcylinders per group\n",ufs->fs_cpg);
fprintf(stderr,"\t\tThis makes %d/%d=%d cg\n",
		ufs->fs_ncyl, ufs->fs_cpg, ufs->fs_ncyl/ufs->fs_cpg
		);
fprintf(stderr,"%u\tinodes per group\n",ufs->fs_ipg);
fprintf(stderr,"\t\tThis makes %d/%d*%d=%d inodes\n",
		ufs->fs_ncyl, ufs->fs_cpg, ufs->fs_ipg,
		ufs->fs_ncyl/ufs->fs_cpg*ufs->fs_ipg
		);

fprintf(stderr,"%u\tblocks per group * fs_frag\n",ufs->fs_fpg);

fprintf(stderr,"\t%u\tnumber of directories\n",ufs->fs_cstotal.cs_ndir);
fprintf(stderr,"\t%u\tnumber of free blocks\n",ufs->fs_cstotal.cs_nbfree);
fprintf(stderr,"\t%u\tnumber of free inodes\n",ufs->fs_cstotal.cs_nifree);
fprintf(stderr,"\t%u\tnumber of free frags\n",ufs->fs_cstotal.cs_nffree);

fprintf(stderr,"%x\tsuper block modified flag\n",ufs->fs_fmod);
fprintf(stderr,"%x\tfile system is clean flag\n",ufs->fs_clean);
fprintf(stderr,"%x\tmounted read-only flag\n",ufs->fs_ronly);
fprintf(stderr,"%x\tcurrently unused flag\n",ufs->fs_flags);

fprintf(stderr,"last mounted on\t%s\n",ufs->fs_fsmnt);
fprintf(stderr,"%x\tlast cg searched\n",ufs->fs_cgrotor);
/*fprintf(stderr,"%x\tlist of fs_cs info buffers\n",ufs->csum *fs_csp[MAXCSBUFS]);*/

fprintf(stderr,"%x\tcyl per cycle in postbl\n",ufs->fs_cpc);

/*fprintf(stderr,"%x\told rotation block list head\n",ufs->fs_opostbl[16][8]);*/
/*fprintf(stderr,"%x\treserved for future constants\n",ufs->fs_sparecon[56]);*/

fprintf(stderr,"%x\t~fs_bmask - for use with quad size\n",ufs->fs_qbmask);
fprintf(stderr,"%x\t~fs_fmask - for use with quad size\n",ufs->fs_qfmask);
fprintf(stderr,"%x\tformat of positional layout tables\n",ufs->fs_postblformat);
fprintf(stderr,"%x\tnumber of rotaional positions\n",ufs->fs_nrpos);
fprintf(stderr,"%u\t(short) rotation block list head\n",ufs->fs_postbloff);
fprintf(stderr,"%u\t(u_char) blocks for each rotation\n",ufs->fs_rotbloff);
fprintf(stderr,"0x%x\tmagic number (should be 0x%x)\n",ufs->fs_magic,FS_MAGIC);
/*fprintf(stderr,"%x\tlist of blocks for each rotation\n",ufs->_char	fs_space[1]);*/
	}

	return(ufs->fs_magic == FS_MAGIC);
}

int main(int argc, char ** argv)
{
	extern char			*optarg;
	extern int 			optind,opterr;
	int 				opt;

	DISK_T				disk;
	int					partition=0;
	char				*name=NULL;
	int					f=0;
	long				osize;
	off_t				start;

	unsigned char		bblock[BBSIZE];
	unsigned char		buffer[BBSIZE]; /* just for testing */
	unsigned char		sblock[SBSIZE]; /* superblock */
	struct fs			*ufs;
	struct dk_label		*label, *plabel;
	short				i, j, k, l;

	struct dinode		*di;
#ifdef _SOLARIS_
	unsigned long		inum = UFSROOTINO;
#else /* SunOS 4.x */
	unsigned long		inum = ROOTINO;
#endif

	unsigned char		*fsb; /* file system block */
	unsigned char		*fsb1; /* file system block, level 1*/
	unsigned char		*fsb2; /* file system block, level 2*/
	unsigned char		*fsdb; /* file system block (for data)*/
	daddr_t				*ibno;
	daddr_t				*ibno1;
	daddr_t				*ibno2;
	int					lbn;

	if ((app = strrchr(argv[0], PATH_SEP))  != NULL)
		app = app+1;
	else
		app = argv[0];

	while ( (opt=getopt(argc,argv,"hrdp:i:Vo:")) != EOF )
	{
		switch(opt)
		{
			case 'r':
				raw=1;
				break;

			case 'd':
				debug=1;
				break;

			case 'o':
				name=optarg;
				break;

			case 'V':
				fprintf(stderr, "%s: %s\n", app, VERS_STR);
				exit(0);
				break;
			case 'i':
				inum=(unsigned long) atol(optarg);
				if (inum<0)
				{
					fprintf(stderr, "%s: invalid inode (%d)\n",
					app, inum);
					exit(1);
				}
				break;
			case 'p':
				if ((*optarg>='a') && (*optarg<='h'))
					partition = (int)(*optarg - 'a');
				else
					partition=atoi(optarg);
				
				if ((partition<0) || (partition>=NDKMAP))
				{
					fprintf(stderr, "%s: invalid parition (%d)\n",
					app, partition);
					exit(1);
				}
				break;
			case 'h':
			default:
				usage();
				break;
		}
	}

	if ( argv[optind] == NULL )
		usage();

	/*
	 * SD(4S)
	 * "raw"  interface: 
	 * I/O requests (such as lseek (2V)) to the SCSI disk must have
	 * an  offset  that  is a multiple of 512 bytes (DEV_BSIZE), or
	 * the driver returns an EINVAL error.  If the transfer  length
	 * is  not  a  multiple  of  512  bytes,  the transfer count is
	 * rounded up by the driver.
	 * ...
	 * A logical block size of 512 bytes is assumed  (and  enforced
	 * on desktop SPARCsystems).
	 */

#ifdef _SOLARIS_
	if ( !raw && ( 0==strncmp ( "/dev/rdsk/", argv[optind], 10 )))
#else /* SunOS 4.x */
	if ( !raw && ( 0==strncmp ( "/dev/r", argv[optind], 6 )))
#endif
	{
		fprintf(stderr,"Warning: assuming raw I/O (%s)\n", argv[optind]);
		raw = 1;
	}	


#ifdef WIN32
    if (INVALID_HANDLE_VALUE == ( disk = CreateFile(
    (LPCTSTR) argv[optind],
    GENERIC_READ,
    0, /* Not shared */
    NULL, /* Not inherited by child processes */
    OPEN_EXISTING,
    FILE_ATTRIBUTE_NORMAL | FILE_FLAG_RANDOM_ACCESS,
    NULL
    )))
#else
    if ( -1 == ( disk = open( argv[optind], O_RDONLY )))
#endif

	{
		merror(argv[optind]);
		exit(1);
	}
	else
	{
		if ( BBSIZE != READ( disk, bblock, BBSIZE ))
			merror("read boot block");

		label = (struct dk_label *)bblock;

		if ( !check_dk_label( label ) )
		{
			fprintf(stderr,"%s: corrupted boot block\n", app);
		}
		/*
		 * Translate partition starting cylinder to offset from 0
		 */
		
		start =
			(off_t) label->dkl_map[partition].dkl_cylno *
			(off_t) label->dkl_nhead *
			(off_t) label->dkl_nsect * 512;


		if ( 0xFFFFFFFF == LSEEK(disk, start, SEEK_SET))
			perror("lseek");

		if ( BBSIZE != READ( disk, buffer, BBSIZE ))
			merror("read boot block");
		
		plabel = (struct dk_label *)buffer; /* trash? */

		if ( SBSIZE != READ( disk, sblock, SBSIZE ))
			merror("can't read superblock in partition");
		
		ufs = (struct fs *)sblock;
		if (debug) fprintf(stderr,"####\tSuperBlock:\n");
		if ( ! check_sblock(ufs))
		{
			fprintf(stderr,"%s: corrupted super block\n", app);
		}

		if (
		( NULL == (fsb = malloc(MAXBSIZE))) ||
		( NULL == (fsb1 = malloc(MAXBSIZE))) ||
		( NULL == (fsb2 = malloc(MAXBSIZE))) ||
		( NULL == (fsdb = malloc(MAXBSIZE)))
		)
		{
			perror("malloc");
			exit(1);
		}

		di = NULL;
		if ( 0!= read_dinode(disk, label, partition, ufs, inum,
				buffer, &di))
		{
			perror("read dinode");
			exit(1);
		}

		if ( name != NULL)
		if ( -1 == (f = open(name, O_TRUNC|O_WRONLY|O_CREAT
#ifdef WIN32
		|O_BINARY
#endif
		, 0666)))
		{
			perror("open");
			exit(1);
		}
			
		if (debug) fprintf(stderr,"####\tinode data:\n");
		osize=0;
		lbn=0;
		for(i=0; i<NDADDR; i++)
		{
			if ( di->di_db[i] == 0 )
				break;

			read_fsblock( disk, label, partition, ufs,
							di, di->di_db[i],
							f, fsdb,
							&osize, &lbn);
		}
		for(i=0; i<NIADDR; i++)
		{
			if (( i == 0 ) && ( di->di_ib[i] != 0 ))
			{
				start = fsdbtoaddr(di->di_ib[i], label, ufs, partition);
			if (debug)
				fprintf(stderr,"\t\tindirect block at 0x%x\n", start);

				if ( 0xFFFFFFFF == LSEEK(disk, start, SEEK_SET))
					perror("lseek idb");


				if ( MAXBSIZE != READ( disk, fsb, MAXBSIZE))
				{
					merror("read idb");
					exit(1);
				}
				for( ibno = (daddr_t *)fsb, j=0;
						j < MAXBSIZE /sizeof(daddr_t);
						ibno ++ , j++ )
				{
					if ( *ibno == 0x00 )
						break;

					read_fsblock( disk, label, partition, ufs,
									di,
#ifdef LITTLEENDIAN
									(swap_word(*ibno)),
#else
									*ibno,
#endif /* LITTLEENDIAN */
									f, fsdb,
									&osize, &lbn);
				}
			}
			else if (( i == 1 ) && ( di->di_ib[i] != 0 ))
			{
				start = fsdbtoaddr(di->di_ib[i], label, ufs, partition);
				if (debug)
				fprintf(stderr,"\t\tindirect block at 0x%x\n", start);

				if ( 0xFFFFFFFF == LSEEK(disk, start, SEEK_SET))
					perror("lseek idb");


				if ( MAXBSIZE != READ( disk, fsb, MAXBSIZE))
				{
					merror("read idb");
					exit(1);
				}
				for( ibno = (daddr_t *)fsb, j=0;
						j < MAXBSIZE /sizeof(daddr_t);
						ibno ++ , j++ )
				{
					if ( *ibno == 0x00 )
						break;

#ifdef LITTLEENDIAN
					SWAP_WORD(*ibno);
#endif /* LITTLEENDIAN */

					start = fsdbtoaddr(*ibno, label, ufs, partition);
					if (debug)
					fprintf(stderr,"\t\tlevel 1 indirect block at 0x%x\n", start);

					if ( 0xFFFFFFFF == LSEEK(disk, start, SEEK_SET))
						perror("lseek idb1");

					if ( MAXBSIZE != READ( disk, fsb1, MAXBSIZE))
					{
						merror("read idb1");
						exit(1);
					}

					for( ibno1 = (daddr_t *)fsb1, k=0;
							k < MAXBSIZE /sizeof(daddr_t);
							ibno1 ++ , k++ )
					{
						if ( *ibno1 == 0x00 )
							break;
						read_fsblock( disk, label, partition, ufs,
									di,
#ifdef LITTLEENDIAN
									(swap_word(*ibno1)),
#else
									*ibno1,
#endif /* LITTLEENDIAN */
									f, fsdb,
									&osize, &lbn);
					}
				}
			}
			else if (( i == 2 ) && ( di->di_ib[i] != 0 ))
			{
				start = fsdbtoaddr(di->di_ib[i], label, ufs, partition);
				if (debug)
				fprintf(stderr,"\t\tindirect block at 0x%x\n", start);

				if ( 0xFFFFFFFF == LSEEK(disk, start, SEEK_SET))
					perror("lseek idb");


				if ( MAXBSIZE != READ( disk, fsb, MAXBSIZE))
				{
					merror("read idb");
					exit(1);
				}
				for( ibno = (daddr_t *)fsb, j=0;
						j < MAXBSIZE /sizeof(daddr_t);
						ibno ++ , j++ )
				{
					if ( *ibno == 0x00 )
						break;

#ifdef LITTLEENDIAN
					SWAP_WORD(*ibno);
#endif /* LITTLEENDIAN */

					start = fsdbtoaddr(*ibno, label, ufs, partition);
					if (debug)
					fprintf(stderr,"\t\tlevel 1 indirect block at 0x%x\n", start);

					if ( 0xFFFFFFFF == LSEEK(disk, start, SEEK_SET))
						perror("lseek idb1");

					if ( MAXBSIZE != READ( disk, fsb1, MAXBSIZE))
					{
						merror("read idb1");
						exit(1);
					}

					for( ibno1 = (daddr_t *)fsb1, k=0;
							k < MAXBSIZE /sizeof(daddr_t);
							ibno1 ++ , k++ )
					{
						if ( *ibno1 == 0x00 )
							break;
#ifdef LITTLEENDIAN
						SWAP_WORD(*ibno1);
#endif /* LITTLEENDIAN */

						start = fsdbtoaddr(*ibno1, label, ufs, partition);
						if (debug)
						fprintf(stderr,"\t\tlevel 2 indirect block at 0x%x\n", start);

						if ( 0xFFFFFFFF == LSEEK(disk, start, SEEK_SET))
							perror("lseek idb2");

						if ( MAXBSIZE != READ( disk, fsb2, MAXBSIZE))
						{
							merror("read idb2");
							exit(1);
						}

						for( ibno2 = (daddr_t *)fsb2, l=0;
								l < MAXBSIZE /sizeof(daddr_t);
								ibno2 ++ , l++ )
						{
							if ( *ibno2 == 0x00 )
								break;

							read_fsblock( disk, label, partition, ufs,
									di,
#ifdef LITTLEENDIAN
									(swap_word(*ibno2)),
#else
									*ibno2,
#endif /* LITTLEENDIAN */
									f, fsdb,
									&osize, &lbn);
						}
					}
				}
			}
		}
		if (fsdb) free(fsdb);

		free( fsb );
		free( fsb1 );
		free( fsb2 );
		if (f!=0) close(f);
		CLOSE(disk);
	}
	exit ( 0 );
}



int read_dinode(
	DISK_T				disk,
	struct dk_label	*label,
	int				partition,
	struct fs		*ufs,
	unsigned long	inum,
	unsigned char	buffer[],

	struct dinode	**di
	)
{
	daddr_t				bno;
	off_t				start;
	off_t				remain;

		/*
		 * macros defined in *fs.h
		 * inode number to file system block address
		 */
		bno = itod( ufs, inum );
		/*
		 * inode number to file system block offset
		 */
		start = (sizeof (struct dinode)) * (itoo( ufs, inum));
		/*
		 * you should scale by fs_fsize, the fragment size.  The reason
		 * for this is that you can then use diskblock addrs to address
		 * fragments rather than blocks as is necessary for files
		 * that have fragments.
		 */
		start += bno * ufs->fs_fsize;

		start += (label->dkl_map[partition].dkl_cylno * \
		label->dkl_nhead * \
		label->dkl_nsect ) * 512;

		if (debug) fprintf(stderr,"####\t%d inode (0x%x):\n", inum, start);

		if ( raw && ( start % DEV_BSIZE != 0 ))
		{
			remain = start % DEV_BSIZE ;
			start -= remain;
		}
		else
		{
			remain = 0L;
		}

		if ( 0xFFFFFFFF == LSEEK(disk, start, SEEK_SET))
		{
			perror("lseek inum");
			return(1);
		}

		if ( BBSIZE != READ( disk, buffer, BBSIZE ))
		{
			merror("read inode");
			return(2);
		}

		if ( *di == NULL )
			*di = (struct dinode *)(buffer + remain);
		else
			memcpy( (char *)*di, buffer + remain, sizeof(struct dinode));

		return (check_dinode(*di));
}


int read_fsblock(
	DISK_T				disk,
	struct dk_label	*label,
	int				partition,
	struct fs		*ufs,

	struct dinode	*di,

	daddr_t			block,
	int				f,
	unsigned char	*fsdb,
	long			*osize,
	int				*lbn
	)
{
	long				fsb_size;
	off_t				start;
	unsigned int		data;
	struct direct		*dir;
	unsigned char		*end;

	unsigned char		buffer[BBSIZE];
	struct dinode		ddi;
	struct dinode		*pddi=&ddi;


	fsb_size = dblksize(ufs, di, *lbn);
	fsb_size = MIN( fsb_size , di->di_size - *osize);
	*osize += (unsigned long)fsb_size;

	start = fsdbtoaddr(block, label, ufs, partition);

	if (debug)
		fprintf(stderr,"\t\t%d\t0x%x (%d, %d/%lld)\n", *lbn, start,
		fsb_size, *osize, di->di_size);

	(*lbn)++ ;

	if ( 0xFFFFFFFF == LSEEK(disk, start, SEEK_SET))
	{
		perror("lseek data");
		return ( 1 );
	}

	/* round up to a 512 boundary for raw I/O */
	data = fsb_size + DEV_BSIZE_ROUND &~ DEV_BSIZE_ROUND;

	data = READ( disk, fsdb, data);
	if ( data == -1 )
	{
		merror("read");
		return ( 1 );
	}

	if (f!=0)
		if ( -1 == write(f, fsdb, fsb_size))
			perror("write");

	switch (di->di_mode & IFMT)
	{
		case IFDIR:
			for(
				end = fsdb + data;
				( fsdb < end ) && (*(u_long *)fsdb != 0);
				fsdb += dir->d_reclen
				)
			{
				dir = (struct  direct *)fsdb;

#ifdef LITTLEENDIAN
				SWAP_WORD(dir->d_ino);
				SWAP_HALF(dir->d_reclen);
				SWAP_HALF(dir->d_namlen);
#endif /* LITTLEENDIAN */

				read_dinode(disk, label, partition, ufs,
						dir->d_ino, buffer, &pddi);

				fprintf(stderr,"%6d %3d %2d %14s (%d)",
				dir->d_ino,
				dir->d_reclen,
				dir->d_namlen,
				dir->d_name,
				DIRSIZ(dir)
				);

				switch (pddi->di_mode & IFMT)
				{
					case IFIFO:
						fprintf(stderr," p ");
						break;
					case IFCHR:
						fprintf(stderr," c ");
						break;
					case IFDIR:
						fprintf(stderr," d ");
						break;
					case IFBLK:
						fprintf(stderr," b ");
						break;
					case IFREG:
						fprintf(stderr," - ");
						break;
					case IFLNK:
						fprintf(stderr," l ");
						break;
					case IFSOCK:
						fprintf(stderr," s ");
						break;
					default:
						fprintf(stderr," ? ");
						break;
				}
						
				fprintf(stderr,"\t%0.4o %2d %4d %4d %8d\n",
				pddi->di_mode & 0x0fff,
				pddi->di_nlink,
				pddi->di_uid,
				pddi->di_gid,
				pddi->di_size);
			}
			break;
			assert( fsdb <= end );

		default:
			break;
	}

	
	return ( 0 );
}
