/*
 * :set tabstop=4
 * ufs_db.c
 * spd@daphne.cps.unizar.es
 * Thu Mar 27 01:27:45 MET 1997
 *
 * UFS debug
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <time.h>

#ifdef _SOLARIS_
#	include <sys/time.h>
#	include <sys/param.h>
#	include <sys/dklabel.h>
#	include <sys/fs/ufs_fs.h>
#	include <sys/fs/ufs_inode.h>
#	include <sys/fs/ufs_fsdir.h>
#endif
#ifdef _SUNOS_ /* SunOS 4.x */
#	include <sys/time.h>
#	include <sys/param.h>
#	include <sun/dklabel.h>
#	include <ufs/fs.h>
#	include <sys/vnode.h>
#	include <ufs/inode.h>
#	include <ufs/fsdir.h>
#endif
#ifdef WIN32
#	include <io.h>
#	include <ctype.h>

#	include <windows.h>

#	include "s_param.h"
#	include "s_types.h"
#	include "s_dklabel.h"
#	include "s_fs.h"
#	include "s_vnode.h"
#	include "s_inode.h"
#	include "s_fsdir.h"

#	define S_IFIFO     0x1000  /* fifo */
#	define S_IFBLK     0x6000  /* block special */
#	define S_IFSOCK    0xC000  /* socket */
#	define S_IFLNK     0xA000  /* symbolic link */
#	ifndef S_ISUID
#		define S_ISUID     0x800   /* set user id on execution */
#		define S_ISGID     0x400   /* set group id on execution */
#	endif  /* S_ISUID */
#	define S_ISVTX     0x200   /* save swapped text even after use */

#endif /* WIN32 */

#include "ufs_db.h"


void ufs_pstat( FILE *f, struct stat *sbp )
{
	int			isdevice=0;

	fprintf(f,"\ttype:");
	switch (sbp->st_mode & S_IFMT)
	{
		case S_IFIFO:
			fprintf(f,"\tnamed pipe (fifo)\n");
			break;
		case S_IFCHR:
			fprintf(f,"\tcharacter special\n");
			isdevice=1;
			break;
		case S_IFDIR:
			fprintf(f,"\tdirectory\n");
			break;
		case S_IFBLK:
			fprintf(f,"\tblock special\n");
			isdevice=1;
			break;
		case S_IFREG:
			fprintf(f,"\tregular\n");
			break;
		case S_IFLNK:
			fprintf(f,"\tsymbolic link\n");
			break;
		case S_IFSOCK:
			fprintf(f,"\tsocket\n");
			break;
		default:
			fprintf(f,"\tunknown\n");
			break;
	}

	if (isdevice)
		fprintf(f, "Device number:%d,%d\n",
				(sbp->st_rdev>>8) & 0xff,
				sbp->st_rdev & 0xff );

	fprintf(f, "Resides on device:%d,%d\n",
			(sbp->st_dev>>8) & 0xff,
			sbp->st_dev & 0xff);

	fprintf(f, "I-node: %d; Links: %d; Size: %ld\n",
			sbp->st_ino, sbp->st_nlink, sbp->st_size);


	fprintf(f, "Owner ID: %d\n", sbp->st_uid );
	fprintf(f, "Group ID: %d\n", sbp->st_gid );

	if ((sbp->st_mode & S_ISUID )== S_ISUID )
		fprintf(f, "Set-user-ID\n");

	if ((sbp->st_mode & S_ISGID )== S_ISGID )
		fprintf(f, "Set-group-ID\n");

	if ((sbp->st_mode & S_ISVTX )== S_ISVTX )
		fprintf(f, "Save swapped text after use\n");

	fprintf(f, "Permissions: %o\n",sbp->st_mode & 0xfff);
	fprintf(f, "Last access:         %s",
		asctime(localtime(&sbp->st_atime )));
	fprintf(f, "Last modification:   %s",
		asctime(localtime(&sbp->st_mtime )));
	fprintf(f, "Last status change:  %s",
		asctime(localtime(&sbp->st_ctime )));

}

void ufs_pdinode(FILE *f, struct dinode *di)
{

	int i;

	fprintf(f,"%x\tmode and type of file\n", di->di_mode);

	switch (di->di_mode & IFMT)
	{
		case IFIFO:
			fprintf(f,"\t\tnamed pipe (fifo)\n");
			break;
		case IFCHR:
			fprintf(f,"\t\tcharacter special\n");
			break;
		case IFDIR:
			fprintf(f,"\t\tdirectory\n");
			break;
		case IFBLK:
			fprintf(f,"\t\tblock special\n");
			break;
		case IFREG:
			fprintf(f,"\t\tregular\n");
			break;
		case IFLNK:
			fprintf(f,"\t\tsymbolic link\n");
			break;
		case IFSOCK:
			fprintf(f,"\t\tsocket\n");
			break;
		default:
			fprintf(f,"\t\tunknown\n");
			break;
	}

	if (di->di_mode & ISUID)
		fprintf(f,"\t\tset user id on execution\n");
	if (di->di_mode & ISGID)
		fprintf(f,"\t\tset group id on execution\n");
	if (di->di_mode & ISVTX)
		fprintf(f,"\t\tsave swapped text even after use\n");
	fprintf(f,"\t\tPermissions: %o\n", di->di_mode & 0x0fff);

	fprintf(f,"%d\tnumber of links to file\n", di->di_nlink);
	fprintf(f,"%d\towner's user id\n", di->di_uid);
	fprintf(f,"%d\towner's group id\n", di->di_gid);
	fprintf(f,"%d\tnumber of bytes in file\n", di->di_size);

	fprintf(f,"\ttime last accessed\t%s", ctime(&di->di_atime));
	fprintf(f,"\tatspare\t%s", ctime(&di->di_ic.ic_atspare));
	fprintf(f,"\ttime last modified\t%s", ctime(&di->di_mtime));
	fprintf(f,"\tmtspare\t%s", ctime(&di->di_ic.ic_mtspare));
	fprintf(f,"\tlast time inode changed\t%s", ctime(&di->di_ctime));
	fprintf(f,"\tctspare\t%s", ctime(&di->di_ic.ic_ctspare));
	fprintf(f,"\t\tic_db[0]=di_rdev\n");

	for(i=0; i<NDADDR; i++)
		fprintf(f,"\t\tdisk block address ic_db[%d]=0x%x\n", i,
		di->di_db[i]
		);
	for(i=0; i<NIADDR; i++)
		fprintf(f,"\t\tindirect block ic_ib[%d]=%x\n", i,
		di->di_ib[i]
		);

	fprintf(f,"%d\tstatus, currently unused\n", di->di_ic.ic_flags);
	fprintf(f,"%d\tblocks actually held\n", di->di_blocks);
	fprintf(f,"%d\tgeneration number\n", di->di_gen);
		/*
		 * XXX - the disk spares were used to avoid changing
		 * the size of the incore inode in a minor release.
		 * Fix for 5.0 release.
		 * Also remove the code in iget, iupdat that clears them.
		 */
#ifdef _SOLARIS_
fprintf(f,"%d\tshadow inode\n", di->di_ic.ic_shadow);
fprintf(f,"%d\tlong EFT version of uid\n", di->di_uid);
fprintf(f,"%d\tlong EFT version of gid\n", di->di_gid);
fprintf(f,"%d\treserved (ic_oeftflag)\n", di->di_ic.ic_oeftflag);
#else
fprintf(f,"%d\tdelayed writes, units=bytes\n", di->di_ic.ic_delaylen);
fprintf(f,"%d\twhere we started delaying\n", di->di_ic.ic_delayoff);
fprintf(f,"%d\twhere to start the next clust\n", di->di_ic.ic_nextrio);
fprintf(f,"%d\tnumber of outstanding writes\n", di->di_ic.ic_writes);
#endif
	
}

void ufs_plabel(FILE *f, struct dk_label *label)
{
	int i;

	fprintf(f,"%s\n",label->dkl_asciilabel);
	/* dkl_pad */
	fprintf(f,"%u\trotations per minute\n",label->dkl_rpm);
	fprintf(f,"%u\tphysical cylinders\n",label->dkl_pcyl);
	fprintf(f,"%u\talternates per cylinder\n",label->dkl_apc);
	fprintf(f,"%x\tobsolete\n",label->dkl_obs1);
	fprintf(f,"%x\tobsolete\n",label->dkl_obs2);
	fprintf(f,"%u\tinterleave factor\n",label->dkl_intrlv);
	fprintf(f,"%u\tdata cylinders\n",label->dkl_ncyl);
	fprintf(f,"%u\talternate cylinders\n",label->dkl_acyl);
	fprintf(f,"%u\theads in this partition\n",label->dkl_nhead);
	fprintf(f,"%u\t512 byte sectors per track",label->dkl_nsect);
	fprintf(f,"\t(This makes a %d Mb disk)\n",
		label->dkl_ncyl * label->dkl_nhead *
		label->dkl_nsect * 512 / 1024 / 1024
		);
	fprintf(f,"%x\tobsolete\n",label->dkl_obs3);
	fprintf(f,"%x\tobsolete\n",label->dkl_obs4);
	for( i=0; i<NDKMAP ; i++)
	{
		if ( (label->dkl_map[i].dkl_nblk !=0 ) ||
			(label->dkl_map[i].dkl_cylno != 0) )
		{
			fprintf(f,"\t%d\tlogical partition\n", i);
			fprintf(f,"\t\t%d\tstarting cylinder (0x%x)\n",
				label->dkl_map[i].dkl_cylno,
				label->dkl_map[i].dkl_cylno *
				label->dkl_nhead *
				label->dkl_nsect * 512
				);
			fprintf(f,"\t\t%d\tblocks (%d b, %d Kb, %d Mb)\n",
				label->dkl_map[i].dkl_nblk,
				label->dkl_map[i].dkl_nblk*512,
				label->dkl_map[i].dkl_nblk*512/1024,
				label->dkl_map[i].dkl_nblk*512/1024/1024
				);
		}
	}
	fprintf(f,"0x%X\tmagic (should be 0x%X)\n",
		label->dkl_magic, DKL_MAGIC);

	fprintf(f,"0x%X\txor checksum of sector\n", label->dkl_cksum);
}

void ufs_psblock(FILE *f, struct fs *ufs)
{

fprintf(f,"%x\tlinked list of file systems\n",ufs->fs_link);
fprintf(f,"%x\tused for incore super blocks\n",ufs->fs_rlink);
fprintf(f,"%x\taddr of super-block in filesys\n",ufs->fs_sblkno);
fprintf(f,"%x\toffset of cyl-block in filesys\n",ufs->fs_cblkno);
fprintf(f,"%x\toffset of inode-blocks in filesys\n",ufs->fs_iblkno);
fprintf(f,"%x\toffset of first data after cg\n",ufs->fs_dblkno);
fprintf(f,"%x\tcylinder group offset in cylinder\n",ufs->fs_cgoffset);
fprintf(f,"%x\tused to calc mod fs_ntrak\n",ufs->fs_cgmask);
fprintf(f,"last time written:\t%s",ctime(&ufs->fs_time));
fprintf(f,"%u\tnumber of blocks in fs (%d b, %d Kb, %d Mb)\n",
			ufs->fs_size,
			ufs->fs_size*fsbtodb(ufs, 1)*512,
			ufs->fs_size*fsbtodb(ufs, 1)*512/1024,
			ufs->fs_size*fsbtodb(ufs, 1)*512/1024/1024
			);
fprintf(f,"%u\tnumber of data blocks in fs (%d b, %d Kb, %d Mb) \n",
			ufs->fs_dsize,
			ufs->fs_dsize*fsbtodb(ufs, 1)*512,
			ufs->fs_dsize*fsbtodb(ufs, 1)*512/1024,
			ufs->fs_dsize*fsbtodb(ufs, 1)*512/1024/1024
			);


fprintf(f,"%u\tnumber of cylinder groups\n",ufs->fs_ncg);
fprintf(f,"%u\tsize of basic blocks in fs\n",ufs->fs_bsize);
fprintf(f,"%u\tsize of frag blocks in fs\n",ufs->fs_fsize);
fprintf(f,"%u\tnumber of frags in a block in fs\n",ufs->fs_frag);
fprintf(f,"%u\tminimum percentage of free blocks\n",ufs->fs_minfree);
fprintf(f,"%u\tnum of ms for optimal next block\n",ufs->fs_rotdelay);
fprintf(f,"%u\tdisk revolutions per second\n",ufs->fs_rps);
fprintf(f,"%x\t``blkoff'' calc of blk offsets\n",ufs->fs_bmask);
fprintf(f,"%x\t``fragoff'' calc of frag offsets\n",ufs->fs_fmask);
fprintf(f,"%x\t``lblkno'' calc of logical blkno\n",ufs->fs_bshift);
fprintf(f,"%x\t``numfrags'' calc number of frags\n",ufs->fs_fshift);
fprintf(f,"%u\tmax number of contiguous blks\n",ufs->fs_maxcontig);
fprintf(f,"%u\tmax number of blks per cyl group\n",ufs->fs_maxbpg);
fprintf(f,"%u\tblock to frag shift\n",ufs->fs_fragshift);
fprintf(f,"%u\tfsbtodb and dbtofsb shift constant\n",ufs->fs_fsbtodb);
fprintf(f,"%u\tactual size of super block\n",ufs->fs_sbsize);
fprintf(f,"%x\tcsum block offset\n",ufs->fs_csmask);
fprintf(f,"%x\tcsum block number\n",ufs->fs_csshift);
fprintf(f,"%u\tvalue of NINDIR\n",ufs->fs_nindir);
fprintf(f,"%u\tvalue of INOPB\n",ufs->fs_inopb);
fprintf(f,"%u\tvalue of NSPF\n",ufs->fs_nspf);
fprintf(f,"%x\toptimization preference, see below\n",ufs->fs_optim);
fprintf(f,"%u\t# sectors\n",ufs->fs_npsect);
fprintf(f,"%u\thardware sector interleave\n",ufs->fs_interleave);
fprintf(f,"%u\tsector 0 skew, per track\n",ufs->fs_trackskew);
fprintf(f,"%x %x\tfile system id\n",ufs->fs_id[0],ufs->fs_id[1]);
fprintf(f,"%x\tblk addr of cyl grp summary area\n",ufs->fs_csaddr);
fprintf(f,"%u\tsize of cyl grp summary area\n",ufs->fs_cssize);
fprintf(f,"%u\tcylinder group size\n",ufs->fs_cgsize);
fprintf(f,"%u\ttracks per cylinder\n",ufs->fs_ntrak);
fprintf(f,"%u\tsectors per track\n",ufs->fs_nsect);
fprintf(f,"%u\tsectors per cylinder (%d*%d=%d)\n",
		ufs->fs_spc,
		ufs->fs_ntrak,ufs->fs_nsect,ufs->fs_ntrak*ufs->fs_nsect
		);
fprintf(f,"%u\tcylinders in file system\n",ufs->fs_ncyl);
fprintf(f,"%u\tcylinders per group\n",ufs->fs_cpg);
fprintf(f,"\t\tThis makes %d/%d=%d cg\n",
		ufs->fs_ncyl, ufs->fs_cpg, ufs->fs_ncyl/ufs->fs_cpg
		);
fprintf(f,"%u\tinodes per group\n",ufs->fs_ipg);
fprintf(f,"\t\tThis makes %d/%d*%d=%d inodes\n",
		ufs->fs_ncyl, ufs->fs_cpg, ufs->fs_ipg,
		ufs->fs_ncyl/ufs->fs_cpg*ufs->fs_ipg
		);

fprintf(f,"%u\tblocks per group * fs_frag\n",ufs->fs_fpg);

fprintf(f,"\t%u\tnumber of directories\n",ufs->fs_cstotal.cs_ndir);
fprintf(f,"\t%u\tnumber of free blocks\n",ufs->fs_cstotal.cs_nbfree);
fprintf(f,"\t%u\tnumber of free inodes\n",ufs->fs_cstotal.cs_nifree);
fprintf(f,"\t%u\tnumber of free frags\n",ufs->fs_cstotal.cs_nffree);

fprintf(f,"%x\tsuper block modified flag\n",ufs->fs_fmod);
fprintf(f,"%x\tfile system is clean flag\n",ufs->fs_clean);
fprintf(f,"%x\tmounted read-only flag\n",ufs->fs_ronly);
fprintf(f,"%x\tcurrently unused flag\n",ufs->fs_flags);

fprintf(f,"last mounted on\t%s\n",ufs->fs_fsmnt);
fprintf(f,"%x\tlast cg searched\n",ufs->fs_cgrotor);

fprintf(f,"%x\tcyl per cycle in postbl\n",ufs->fs_cpc);


fprintf(f,"%x\t~fs_bmask - for use with quad size\n",ufs->fs_qbmask);
fprintf(f,"%x\t~fs_fmask - for use with quad size\n",ufs->fs_qfmask);
fprintf(f,"%x\tformat of positional layout tables\n",ufs->fs_postblformat);
fprintf(f,"%x\tnumber of rotaional positions\n",ufs->fs_nrpos);
fprintf(f,"%u\t(short) rotation block list head\n",ufs->fs_postbloff);
fprintf(f,"%u\t(u_char) blocks for each rotation\n",ufs->fs_rotbloff);
fprintf(f,"0x%x\tmagic number (should be 0x%x)\n",ufs->fs_magic,FS_MAGIC);

}

