/*
 * :set tabstop=4
 * ufs_file.c
 * spd@daphne.cps.unizar.es
 * Thu Mar 20 00:18:51 MET 1997
 *
 * UFS file/dir high level I/O
 * SunOS 4.x, 5.x (sparc) and WinNT (ix86)
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#ifdef _SOLARIS_
#	include <unistd.h>
#	include <sys/time.h>
#	include <sys/param.h>
#   include <sys/dklabel.h>
#   include <sys/fs/ufs_fs.h>
#	include <sys/fs/ufs_inode.h>
#	include <sys/fs/ufs_fsdir.h>
#endif
#ifdef _SUNOS_
#	include <unistd.h>
#	include <sys/time.h>
#	include <sys/param.h>
#	include <sun/dklabel.h>
#   include <ufs/fs.h>
#	include <sys/vnode.h>
#	include <ufs/inode.h>
#	include <ufs/fsdir.h>
#endif
#ifdef WIN32
#   include <io.h>
#   include <ctype.h>

#   include <windows.h>

#	include "s_param.h"
#	include "s_types.h"
#	include "s_dklabel.h"
#	include "s_fs.h"
#	include "s_vnode.h"
#	include "s_inode.h"
#	include "s_fsdir.h"

#	ifndef S_ISDIR
#		define S_ISDIR(mode)   ((mode&S_IFMT) == S_IFDIR)
#	endif
#endif

#include "sun_ufs.h"
#include "sun_io.h"
#include "ufs_io.h"

#ifdef WIN32 
DWORD _n_read;
#endif

UFS_FILE	_ufs_fileb[UFS_MAX_FILE];

static long read_fsblock(
	UFS_FS			*fs,
	struct dinode	*di,
	daddr_t			block,
	unsigned char	*fsdb,
	long			*osize,
	int				*lbn
	);


/************************************************************************/

UFS_FILE *ufs_fopen( char *name, char *type )
{
	register UFS_FILE	*p;
	register int		i;
	unsigned long		inum;
	unsigned char		buffer[BBSIZE];
	UFS_FS				*fs;

	if ( NULL == (fs = get_default_fs()))
		return 0;

	if ( 0 == ( inum=path_to_inum(name)))
	{
		return(NULL);
	}


	for( p = &_ufs_fileb[0], i=UFS_MAX_FILE; i-- ; p++)
	{
		if (!p->fs)
		{
			if ( NULL == (p->base = malloc(MAXBSIZE)))
			{
				ufs_errno=UFS_EAGAIN;
				return(NULL);
			}
			if ( NULL == (p->di = malloc(sizeof(struct dinode))))
			{
				free(p->base);
				ufs_errno=UFS_EAGAIN;
				return(NULL);
			}
			if ( NULL == (p->fsb = malloc(MAXBSIZE)))
			{
				free(p->base);
				free(p->di);
				ufs_errno=UFS_EAGAIN;
				return(NULL);
			}

			if ( 0 != read_dinode( fs, inum, buffer, &(p->di) ))
			{
				ufs_fclose( p );
				return(NULL);
			}
			p->fs = fs;
			p->cnt = 0;
			p->osize = 0;
			p->lbn = 0;
			p->ilbn = 0;
			p->ptr = p->base;
			p->ibno = (((daddr_t *)(p->fsb)) + MAXBSIZE/sizeof(daddr_t));
			return(p);
		}
	}
	ufs_errno = UFS_EMFILE;
	return(NULL);
}

int ufs_fclose( UFS_FILE *stream )
{
	if ( stream->fsb )
		free( stream->fsb );
	if ( stream->di )
		free( stream->di );
	if ( stream->base )
		free( stream->base );

	stream->fs=NULL;

	return(0);
}

size_t ufs_fread(void *ptr, size_t size, size_t nitems, UFS_FILE *f)
{
	size_t			n, m;
	size_t			r=0;
	daddr_t			block;
	off_t			start;

	if ( 0 == ( n = size * nitems ))
		return 0;
	/* disk eof test here */

	ufs_errno = UFS_ENOERR;

	while ( n > 0 )
	{
		m = MIN( f->cnt, n );

		if ( m > 0 )
		{
			/*
			 * enough bytes in buffer
			 */
		
			if (ptr)
			{
				memcpy( ptr, f->ptr, m );
				((char *)ptr) += m;
			}
			f->ptr += m;
			f->cnt -= m;
			n -= m;
			r += m;
		}
		else
		{
			/*
			 * Get next fs block
			 */
			block = 0;

			if ( f->lbn < NDADDR )
			{
				block = f->di->di_db[f->lbn];
			}
			else if ( f->ilbn < NIADDR )
			{
				if ( f->ibno ==
				(((daddr_t *)(f->fsb)) + MAXBSIZE/sizeof(daddr_t)))
				{
					/*
					 * get next indirect block
					 */
					if ( f->di->di_ib[f->ilbn] != 0 )
					{
						start = fsdbtoaddr(
							f->di->di_ib[f->ilbn],
							f->fs->label,
							f->fs->ufs,
							f->fs->partition);

						if ( start == LSEEK(f->fs->file, start, SEEK_SET))
						{
							if (
								MAXBSIZE ==
								READ( f->fs->file, f->fsb, MAXBSIZE)
								)
							{
								/*
								 * so now fsb holds the ib addrs
								 */
								f->ibno = (daddr_t *)(f->fsb);
								block =
#ifdef LITTLEENDIAN
									(swap_word(*(f->ibno)));
#else
									*(f->ibno);
#endif /* LITTLEENDIAN */

								f->ilbn ++;
							}
							else
								ufs_errno = UFS_EREAD;
						}
						else
							ufs_errno = UFS_ESEEK;
					}
				}
				else
				{
					f->ibno ++;
					block =
#ifdef LITTLEENDIAN
						(swap_word(*(f->ibno)));
#else
						*(f->ibno);
#endif /* LITTLEENDIAN */

				}
			}

			if ( block == 0 )
			{
				if (ufs_errno == UFS_ENOERR)
					f->ilbn = NIADDR; /* eof */
			}
			else
			{
				f->ptr = f->base;
				f->cnt =
					(size_t) read_fsblock(
						f->fs,
						f->di,
						block,
						f->base,
						&f->osize,
						&f->lbn
						);
				if ( f->osize >= f->di->di_size )
					f->ilbn = NIADDR; /* eof */
			}

			if (( f->ilbn == NIADDR ) && (f->cnt==0))
				return( r/size );
		}
	}
	return( r/size );
}

/************************************************************************/

UFS_DIR *ufs_opendir( char *name )
{
	struct stat	st;

	if ( -1 == ufs_stat(name, &st))
		return(NULL);

	if ( S_ISDIR( st.st_mode ))
	{
		return( (UFS_DIR *) ufs_fopen( name, "r"));
	}
	else
	{
		ufs_errno = UFS_ENOTDIR;
		return(NULL);
	}
}

int ufs_closedir( UFS_DIR *dir )
{
	return( ufs_fclose ( (UFS_FILE *) dir ));
}

struct direct *ufs_readdir( UFS_DIR *f )
{
	struct direct	*dir=NULL;
	size_t			n;
	
	while ( dir == NULL )
	{
		if ( f->cnt == 0)
		{
			if ( f->ilbn == NIADDR )
			{
				/*
				 * eof
				 */
				return NULL;
			}
			n = ufs_fread( NULL, 1, MAXBSIZE, f );
			f->ptr = f->base;
			f->cnt = n;
		}
		if ( f->cnt == 0)
		{	
			return NULL;
		}

		dir = (struct  direct *)(f->ptr);

#ifdef LITTLEENDIAN
		SWAP_WORD(dir->d_ino);
		SWAP_HALF(dir->d_reclen);
		SWAP_HALF(dir->d_namlen);
#endif /* LITTLEENDIAN */

		f->ptr += dir->d_reclen;
		f->cnt -= dir->d_reclen;

		if ( dir->d_ino == 0 )
			dir = NULL;
	}

	return dir;
}

/************************************************************************/

static long read_fsblock(
	UFS_FS			*fs,
	struct dinode	*di,
	daddr_t			block,
	unsigned char	*fsdb,
	long			*osize,
	int				*lbn
	)
{
	long			fsb_size;
	off_t			start;
	unsigned int	data;


	fsb_size = dblksize( fs->ufs, di, *lbn);
	fsb_size = MIN( fsb_size , di->di_size - *osize);
	*osize += (unsigned long)fsb_size;
	start = fsdbtoaddr(block, fs->label, fs->ufs, fs->partition);
	(*lbn)++;

	if ( start != LSEEK(fs->file, start, SEEK_SET) )
	{
		ufs_errno = UFS_ESEEK;
		return(0);
	}

	/* round up to a 512 boundary for raw I/O */
	data = fsb_size + DEV_BSIZE_ROUND &~ DEV_BSIZE_ROUND;

	data = READ( fs->file, fsdb, data);
	if ( data == -1 )
	{
		ufs_errno = UFS_EREAD;
		return(0);
	}

	return(fsb_size);
}

/************************************************************************/




