/*
 * :set tabstop=4
 * ufs_fs.c
 * spd@daphne.cps.unizar.es
 * Thu Mar 20 00:18:51 MET 1997
 *
 * ufs mounting
 * SunOS 4.x 5.x, WinNT
 *
 * All fs related calculations are macro-based, so you don't need
 * external libraries
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#ifdef _SOLARIS_
#	include <unistd.h>
#	include <sys/param.h>
#   include <sys/dklabel.h>
#   include <sys/fs/ufs_fs.h>
#endif
#ifdef _SUNOS_
#	include <unistd.h>
#	include <sys/param.h>
#	include <sun/dklabel.h>
#   include <ufs/fs.h>
#endif
#ifdef WIN32
#   include <io.h>
#   include <ctype.h>

#   include <windows.h>

#	include "s_param.h"
#   include "s_types.h"
#	include "s_dklabel.h"
#	include "s_fs.h"
#	ifndef S_ISDIR
#		define S_ISDIR(mode)   ((mode&S_IFMT) == S_IFDIR)
#	endif
#endif

#include "sun_ufs.h"
#include "sun_io.h"
#include "ufs_io.h"

/*
 * little ego
 */
static char *__ident="@(#)libufs 0.1.1 (c) SPDsoft, Apr 12, 1997";

/*
 * ANSI:  If no explicit initialization is performed, static
 * objects are implicitly initialized to zero. 
 */


UFS_DISK	_ufs_diskb[UFS_MAX_DISKS];
UFS_FS		_ufs_fsb[UFS_MAX_FSS];
static UFS_FS	*_current_fs=NULL;

int  ufs_errno=UFS_ENOERR;
char *ufs_sys_errlist[]=
	{
		"No error",
		"Bad checksum",
		"Bad magic number",
		"Read error",
		"Empty partition",
		"Bad partition",
		"Seek error",
		"Resource temporarily unavailable",
		"Too many open files",
		"Not a directory",
		"Filename too long",
		"No such file or directory"
	};

static int check_dk_label(struct dk_label *label);
static int check_sblock(struct fs *ufs);


#ifdef WIN32 
DWORD _n_read;
#endif


UFS_DISK *ufs_open_disk( char *name )
{
	register UFS_DISK *p;
	register int i;

	for( p = &_ufs_diskb[0], i=UFS_MAX_DISKS; i-- ; p++)
	{
		if (!p->file)
		{
			if ( NULL == (p->label = malloc(BBSIZE)))
			{
				ufs_errno=UFS_EAGAIN;
				return(NULL);
			}
#ifdef WIN32 
			if (INVALID_HANDLE_VALUE == ( p->file = CreateFile(
			(LPCTSTR) name,
			GENERIC_READ,
			0, /* Not shared */
			NULL, /* Not inherited by child processes */
			OPEN_EXISTING,
			FILE_ATTRIBUTE_NORMAL | FILE_FLAG_RANDOM_ACCESS,
			NULL
			)))
#else
			if ( -1 == ( p->file = open( name, O_RDONLY )))
#endif
			{
				p->file=(UFS_DISK_T)0L;
				return(NULL);
			}
			else
			{
				if ( BBSIZE != READ( p->file, (void *)p->label, BBSIZE ))
				{
					ufs_errno = UFS_EREAD;
					ufs_close_disk( p );
					return (NULL);
				}
				else
				{
					if (!check_dk_label( p->label ))
					{
						ufs_close_disk( p );
						return (NULL);
					}
				}
				return(p);
			}
		}
	}
	ufs_errno = UFS_EMFILE;
	return(NULL);
}

int ufs_close_disk( UFS_DISK *disk )
{
	int		err;

	err = CLOSE( disk->file );

	disk->file=(UFS_DISK_T)0L;
	if (disk->label)
		free( disk->label );
	
	return(err);
}

/************************************************************************/

UFS_FS   *ufs_mount( UFS_DISK *disk, int partition, int flags )
{
	register UFS_FS		*p;
	register int		i;
	off_t				start;


	if ( ufs_empty_partition( disk, partition ))
	{
		ufs_errno = UFS_EEMPTY;
		return (NULL);
	}

	if ((!( flags & UFS_FMOUNTALL ) && ((partition==2)||(partition==1)))||
	(partition<0) || (partition >7))
	{
		ufs_errno = UFS_EBADPART;
		return (NULL);
	}

	for( p = &_ufs_fsb[0], i=UFS_MAX_FSS; i-- ; p++)
	{
		if (!p->file)
		{
			if ( NULL == (p->ufs = malloc(SBSIZE)))
			{
				ufs_errno=UFS_EAGAIN;
				return(NULL);
			}
			if ( NULL == (p->cwd = malloc(MAXPATHLEN)))
			{
				free(p->ufs);
				ufs_errno=UFS_EAGAIN;
				return(NULL);
			}
			p->file = disk->file;
			p->partition = partition;
			p->label = disk->label;

			p->inum = UFS_ROOTINO;
			strcpy(p->cwd, "/");

			/*
			 * Translate partition starting cylinder to offset from 0
			 */

			start =
				(off_t) p->label->dkl_map[p->partition].dkl_cylno *
				(off_t) p->label->dkl_nhead *
				(off_t) p->label->dkl_nsect * 512
				+ BBSIZE;


			if ( start != LSEEK(p->file, start, SEEK_SET))
			{
				ufs_errno = UFS_ESEEK;
				ufs_umount( p );
				return (NULL);
			}
			else if ( SBSIZE != READ( p->file, (void *)p->ufs, SBSIZE ))
			{
				ufs_errno = UFS_EREAD;
				ufs_umount( p );
				return (NULL);
			}
			else
			{
				if (!check_sblock( p->ufs ))
				{
					ufs_errno = UFS_EMAGIC;
					ufs_umount( p );
					return (NULL);
				}
			}
			if ( _current_fs==NULL )
				_current_fs=p;
			return(p);
		}
	}
	ufs_errno = UFS_EMFILE;
	return(NULL);
}

int  ufs_umount(UFS_FS *fs)
{
	fs->file=(UFS_DISK_T)0L;
	if (fs->ufs)
		free( fs->ufs );
	if (fs->cwd)
		free( fs->cwd );

	if ( _current_fs == fs )
		_current_fs = NULL;

	return(0);
}

UFS_FS   *get_default_fs( void )
{
	return _current_fs;
}
void set_default_fs( UFS_FS *fs )
{
	_current_fs = fs;
}

char *ufs_getwd(char *pathname)
{
	if ( !_current_fs )
	{
		strcpy(pathname, "No current fs");
		return(NULL);
	}
	else
	{
		strncpy( pathname, _current_fs->cwd, MAXPATHLEN);
		return( pathname );
	}
}

int      ufs_chdir(char *ipath)
{
	unsigned long inum;
	struct stat		st;

	char		path[MAXPATHLEN];
	char		ncwd[MAXPATHLEN];
	char		*argt;
	char		*p;
	int			rem;

	if ( 0 == (inum = path_to_inum(ipath)))
		return(-1);

	if ( -1 == ufs_istat(inum, &st))
		return(-1);
	
	if ( S_ISDIR( st.st_mode ))
	{
		strncpy(path, ipath, MAXPATHLEN);
/*
 * assertion:
 *		cwd : "/" or ("/name")* where name != "." or ".."
 */
		if ( *path == '/' )
			ncwd[0]=0x00;
		else
		{
			if ( 0==strcmp( _current_fs->cwd, "/") )
				ncwd[0]=0x00;
			else
				strncpy(ncwd, _current_fs->cwd, MAXPATHLEN);
		}
		rem = MAXPATHLEN - strlen(ncwd);
	
		for(
			argt = strtok( path, "/" );
			argt != NULL;
			argt = strtok( NULL, "/" )
		)
		{
			if ( 0==strcmp( argt, ".") )
				continue;
		
			if ( 0==strcmp( argt, ".."))
			{
				if ( p = (strrchr(ncwd, (int)'/')))
				{
					*p = 0x00;
					rem = MAXPATHLEN - strlen(ncwd);
				}
			}
			else
			{
				rem -= 1 + strlen(argt);
				if ( rem > 0 )
				{
					strcat( ncwd, "/" );
					strcat( ncwd, argt );
				}
				else
				{
					ufs_errno = UFS_ENAMETOOLONG;
					return -1;
				}
			}
		}
		if ( *ncwd == 0x00 )
			strcpy(_current_fs->cwd, "/");
		else
			strncpy(_current_fs->cwd, ncwd, MAXPATHLEN);

		_current_fs->inum = inum;
		return 0;
	}
	else
	{
		ufs_errno=UFS_ENOTDIR;
		return (-1);
	}
}


/************************************************************************/
/************************************************************************/

static int check_dk_label(struct dk_label *label)
{
	short				i;
	unsigned short		sum, *word;

	for (
		word = (unsigned short *)label, sum=0;
		word < ((unsigned short *)label + sizeof(*label)/sizeof(short) -1);
		word ++
		)
		sum ^= *word;

#ifdef LITTLEENDIAN

	SWAP_HALF(label->dkl_rpm);
	SWAP_HALF(label->dkl_pcyl);
	SWAP_HALF(label->dkl_apc);
	SWAP_HALF(label->dkl_intrlv);
	SWAP_HALF(label->dkl_ncyl);
	SWAP_HALF(label->dkl_acyl);
	SWAP_HALF(label->dkl_nhead);
	SWAP_HALF(label->dkl_nsect);
	for( i=0; i<NDKMAP ; i++)
	{
		SWAP_WORD(label->dkl_map[i].dkl_cylno);
		SWAP_WORD(label->dkl_map[i].dkl_nblk);
	}
	SWAP_HALF(label->dkl_magic);

#endif /* LITTLEENDIAN */

	ufs_errno = UFS_ENOERR;
	if ( sum != label->dkl_cksum )
		ufs_errno = UFS_ECSUM;
	if ( DKL_MAGIC != label->dkl_magic )
		ufs_errno = UFS_EMAGIC;

	return ( ufs_errno == UFS_ENOERR );
}
/************************************************************************/

static int		check_sblock(struct fs *ufs)
{
#ifdef LITTLEENDIAN
	SWAP_WORD(ufs->fs_sblkno);
	SWAP_WORD(ufs->fs_cblkno);
	SWAP_WORD(ufs->fs_iblkno);
	SWAP_WORD(ufs->fs_dblkno);
	SWAP_WORD(ufs->fs_cgoffset);
	SWAP_WORD(ufs->fs_cgmask);

	SWAP_WORD(ufs->fs_time);

	SWAP_WORD(ufs->fs_size);
	SWAP_WORD(ufs->fs_dsize);

	SWAP_WORD(ufs->fs_ncg);
	SWAP_WORD(ufs->fs_bsize);
	SWAP_WORD(ufs->fs_fsize);
	SWAP_WORD(ufs->fs_frag);
	SWAP_WORD(ufs->fs_minfree);
	SWAP_WORD(ufs->fs_rotdelay);
	SWAP_WORD(ufs->fs_rps);
	SWAP_WORD(ufs->fs_bmask);
	SWAP_WORD(ufs->fs_fmask);
	SWAP_WORD(ufs->fs_bshift);
	SWAP_WORD(ufs->fs_fshift);
	SWAP_WORD(ufs->fs_maxcontig);
	SWAP_WORD(ufs->fs_maxbpg);
	SWAP_WORD(ufs->fs_fragshift);
	SWAP_WORD(ufs->fs_fsbtodb);
	SWAP_WORD(ufs->fs_sbsize);
	SWAP_WORD(ufs->fs_csmask);
	SWAP_WORD(ufs->fs_csshift);
	SWAP_WORD(ufs->fs_nindir);
	SWAP_WORD(ufs->fs_inopb);
	SWAP_WORD(ufs->fs_nspf);
	SWAP_WORD(ufs->fs_optim);
	SWAP_WORD(ufs->fs_npsect);
	SWAP_WORD(ufs->fs_interleave);
	SWAP_WORD(ufs->fs_trackskew);
	SWAP_WORD(ufs->fs_id[0]);
	SWAP_WORD(ufs->fs_id[1]);
	SWAP_WORD(ufs->fs_csaddr);
	SWAP_WORD(ufs->fs_cssize);
	SWAP_WORD(ufs->fs_cgsize);
	SWAP_WORD(ufs->fs_ntrak);
	SWAP_WORD(ufs->fs_nsect);
	SWAP_WORD(ufs->fs_spc);
	SWAP_WORD(ufs->fs_ncyl);
	SWAP_WORD(ufs->fs_cpg);
	SWAP_WORD(ufs->fs_ipg);

	SWAP_WORD(ufs->fs_fpg);

	SWAP_WORD(ufs->fs_cstotal.cs_ndir);
	SWAP_WORD(ufs->fs_cstotal.cs_nbfree);
	SWAP_WORD(ufs->fs_cstotal.cs_nifree);
	SWAP_WORD(ufs->fs_cstotal.cs_nffree);

	SWAP_WORD(ufs->fs_cgrotor);

	SWAP_WORD(ufs->fs_cpc);

	SWAP_WORD(ufs->fs_qbmask.val[0]);
	SWAP_WORD(ufs->fs_qbmask.val[1]);
	SWAP_WORD(ufs->fs_qfmask.val[0]);
	SWAP_WORD(ufs->fs_qfmask.val[1]);
	SWAP_WORD(ufs->fs_postblformat);
	SWAP_WORD(ufs->fs_nrpos);
	SWAP_WORD(ufs->fs_postbloff);
	SWAP_WORD(ufs->fs_rotbloff);
	SWAP_WORD(ufs->fs_magic);

#endif /* LITTLEENDIAN */

	return(ufs->fs_magic == FS_MAGIC);
}
