/*
 * :set tabstop=4
 * ufs_io.c
 * spd@daphne.cps.unizar.es
 * Thu Mar 20 00:18:51 MET 1997
 *
 * SunOS 4.x, 5.x and WinNT
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#ifdef _SOLARIS_
#	include <unistd.h>
#	include <sys/time.h>
#	include <sys/param.h>
#   include <sys/dklabel.h>
#   include <sys/fs/ufs_fs.h>
#	include <sys/fs/ufs_inode.h>
#	include <sys/fs/ufs_fsdir.h>
#endif
#ifdef _SUNOS_
#	include <unistd.h>
#	include <sys/time.h>
#	include <sys/param.h>
#	include <sun/dklabel.h>
#   include <ufs/fs.h>
#	include <sys/vnode.h>
#	include <ufs/inode.h>
#	include <ufs/fsdir.h>
#endif
#ifdef WIN32
#   include <io.h>
#   include <ctype.h>

#   include <windows.h>

#	include "s_param.h"
#	include "s_types.h"
#	include "s_dklabel.h"
#	include "s_fs.h"
#	include "s_vnode.h"
#	include "s_inode.h"
#	include "s_fsdir.h"
#endif

#include "sun_ufs.h"
#include "sun_io.h"
#include "ufs_io.h"


#ifdef WIN32 
DWORD _n_read;
#endif

static int check_dinode(struct dinode *di);
static int scan_block(
	UFS_FS			*fs,
	struct dinode	*di,
	daddr_t			block,
	unsigned char	*fsdb,
	long			*osize,
	int				*lbn,

	char			*name,
	unsigned long	*inum
	);



unsigned long path_to_inum( const char *ipath )
{
	unsigned long	inum;
	unsigned long	finum=0;
	UFS_FS			*fs;
	char			*argt;

	unsigned char	buffer[BBSIZE];
	unsigned char	fsdb[MAXBSIZE];
	unsigned char	fsb[MAXBSIZE];
	char			path[MAXPATHLEN];

	struct dinode	*di=NULL;

	int				lbn;
	long			osize;
	daddr_t			*ibno;

	off_t			start;

	int				i;
	int				j;


	strncpy(path, ipath, MAXPATHLEN);

	if ( NULL == (fs = get_default_fs()))
		return 0;

	if ( *path == '/' )
		inum = UFS_ROOTINO;
	else
		inum = fs->inum;

	for(
		argt = strtok( path, "/" );
		argt != NULL;
		argt = strtok( NULL, "/" )
		)
	{
		if ( 0 != read_dinode( fs, inum, buffer, &di ))
			return(0);

		if ( (di->di_mode & IFMT) != IFDIR )
		{
			ufs_errno = UFS_ENOTDIR;
			return(0);
		}

		lbn = 0;
		osize = 0;

		for(i=0; i<NDADDR; i++)
		{
			if ( di->di_db[i] == 0 )
				break;

			if ( 0 != scan_block( fs, di, di->di_db[i],
			fsdb, &osize, &lbn, argt, &finum))
				break;

			if ( finum != 0 )
				break;
		}

		if ( i==NDADDR )
		for(i=0; i<NIADDR; i++)
		{
			if ( di->di_ib[i] != 0 )
			{
				start = fsdbtoaddr(
					di->di_ib[i], fs->label, fs->ufs, fs->partition);

				if ( start != LSEEK(fs->file, start, SEEK_SET))
				{
					ufs_errno = UFS_ESEEK;
					break;
				}
				if ( MAXBSIZE != READ( fs->file, fsb, MAXBSIZE))
				{
					ufs_errno = UFS_EREAD;
					break;
				}
				for( ibno = (daddr_t *)fsb, j=0;
					j < MAXBSIZE /sizeof(daddr_t);
					ibno ++ , j++ )
				{
					if ( *ibno == 0x00 )
						break;

					if ( 0 != scan_block( fs, di,
#ifdef LITTLEENDIAN
						swap_word(*ibno),
#else
						*ibno,
#endif /* LITTLEENDIAN */
						fsdb, &osize, &lbn, argt, &finum))
					break;

					if ( finum != 0 )
						break;
				}
			}
			if ( (ufs_errno != UFS_ENOERR ) || ( finum != 0 ))
				break;
		}
		if ( finum == 0 )
		{
			ufs_errno = UFS_ENOENT;
			return(0);
		}
		inum = finum;
	}

	return inum;
}

int ufs_stat(const char *path, struct stat *buf)
{
	unsigned long	inum;
	if ( 0 == (inum = path_to_inum(path)))
		return(-1);

	return( ufs_istat( inum, buf ));
}

int ufs_istat(unsigned long inum, struct stat *buf)
{
	UFS_FS			*fs;

	unsigned char	buffer[BBSIZE];
	struct dinode	*di=NULL;

	if ( NULL == (fs = get_default_fs()))
		return -1;

	if ( 0 != read_dinode( fs, inum, buffer, &di ))
		return -1;

	buf->st_mode = di->di_mode;
#ifdef WIN32
	buf->st_ino = (unsigned short) inum;
#else
	buf->st_ino = inum;
#endif
	buf->st_nlink = di->di_nlink;
	buf->st_uid = di->di_uid;
	buf->st_gid = di->di_gid;
	buf->st_size = di->di_size;
	buf->st_atime = di->di_atime;
	buf->st_mtime = di->di_mtime;
	buf->st_ctime = di->di_ctime;

	return(0);
}

/************************************************************************/

int read_dinode(
	UFS_FS			*fs,
	unsigned long	inum,
	unsigned char	buffer[],
	struct dinode	**di
	)
{
	daddr_t				bno;
	off_t				start;
	off_t				remain;

		/*
		 * macros defined in *fs.h
		 * inode number to file system block address
		 */
		bno = itod( fs->ufs, inum );
		/*
		 * inode number to file system block offset
		 */
		start = (sizeof (struct dinode)) * (itoo(fs->ufs, inum));
		/*
		 * you should scale by fs_fsize, the fragment size.  The reason
		 * for this is that you can then use diskblock addrs to address
		 * fragments rather than blocks as is necessary for files
		 * that have fragments.
		 */
		start += bno * fs->ufs->fs_fsize;

		start += (fs->label->dkl_map[fs->partition].dkl_cylno * \
		fs->label->dkl_nhead * \
		fs->label->dkl_nsect ) * 512;

		if ( start % DEV_BSIZE != 0 )
		{
			remain = start % DEV_BSIZE ;
			start -= remain;
		}
		else
		{
			remain = 0L;
		}

		if ( start != LSEEK(fs->file, start, SEEK_SET) )
		{
			ufs_errno = UFS_ESEEK;
			return(1);
		}

		if ( BBSIZE != READ(fs->file, buffer, BBSIZE) )
		{
			ufs_errno = UFS_EREAD;
			return(2);
		}

		if ( *di == NULL )
			*di = (struct dinode *)(buffer + remain);
		else
			memcpy( (char *)*di, buffer + remain, sizeof(struct dinode));

		return (check_dinode(*di));
}


/************************************************************************/

static int check_dinode(struct dinode *di)
{
#ifdef LITTLEENDIAN
	int i;

	SWAP_HALF(di->di_mode);

	SWAP_HALF(di->di_nlink);
	SWAP_HALF(di->di_uid);
	SWAP_HALF(di->di_gid);
	SWAP_WORD(di->di_size);

	SWAP_WORD(di->di_atime);
	SWAP_WORD(di->di_ic.ic_atspare);
	SWAP_WORD(di->di_mtime);
	SWAP_WORD(di->di_ic.ic_mtspare);
	SWAP_WORD(di->di_ctime);
	SWAP_WORD(di->di_ic.ic_ctspare);

	for(i=0; i<NDADDR; i++)
		SWAP_WORD(di->di_db[i]);

	for(i=0; i<NIADDR; i++)
		SWAP_WORD(di->di_ib[i]);

	SWAP_WORD(di->di_ic.ic_flags);
	SWAP_WORD(di->di_blocks);
	SWAP_WORD(di->di_gen);

#endif /* LITTLEENDIAN */

	return(0);
}



static int scan_block(
	UFS_FS			*fs,
	struct dinode	*di,
	daddr_t			block,
	unsigned char	*fsdb,
	long			*osize,
	int				*lbn,

	char			*name,
	unsigned long	*inum
	)
{
	long			fsb_size;
	off_t			start;
	unsigned int	data;
	struct direct	*dir;
	unsigned char	*end;


	fsb_size = dblksize( fs->ufs, di, *lbn);
	fsb_size = MIN( fsb_size , di->di_size - *osize);
	*osize += (unsigned long)fsb_size;
	start = fsdbtoaddr(block, fs->label, fs->ufs, fs->partition);
	(*lbn)++;

	if ( start != LSEEK(fs->file, start, SEEK_SET) )
	{
		ufs_errno = UFS_ESEEK;
		return(1);
	}

	/* round up to a 512 boundary for raw I/O */
	data = fsb_size + DEV_BSIZE_ROUND &~ DEV_BSIZE_ROUND;

	data = READ( fs->file, fsdb, data);
	if ( data == -1 )
	{
		ufs_errno = UFS_EREAD;
		return(2);
	}

	*inum = 0;
	for(
		end = fsdb + data;
		fsdb < end ;
		fsdb += dir->d_reclen
		)
	{
		dir = (struct  direct *)fsdb;
#ifdef LITTLEENDIAN
		SWAP_WORD(dir->d_ino);
		SWAP_HALF(dir->d_reclen);
		SWAP_HALF(dir->d_namlen);
#endif /* LITTLEENDIAN */
		if ( dir->d_ino == 0 )
			continue;
	
		if (0 == strcmp(name, dir->d_name))
		{
			*inum = dir->d_ino;
			break;
		}
	}
	return(0);
}
/************************************************************************/
